"""Boosting as described by David et al. (2007).

Versions
--------
Stored in ``algorithm_version`` attribute (see docstring)

Profiling
---------
ds = datasets._get_continuous()
y = ds['y']
x1 = ds['x1']
x2 = ds['x2']

%prun -s cumulative res = boosting(y, x1, 0, 1)

"""
from __future__ import annotations

from dataclasses import dataclass, field, fields
from functools import cached_property, reduce
import inspect
from itertools import chain, repeat
from math import ceil
from operator import mul
import time
from typing import Any, Callable, List, Literal, Optional, Union, Tuple, Sequence
import warnings

import numba
import numpy as np

from .._config import CONFIG
from .._data_obj import Case, Dataset, Dimension, SourceSpaceBase, NDVar, CategorialArg, NDVarArg, dataobj_repr
from .._exceptions import OldVersionError
from .._ndvar.ndvar import _concatenate_values, convolve_jit, set_connectivity, set_parc
from .._utils import PickleableDataClass, deprecate_ds_arg, user_activity
from .shared import PredictorData, DeconvolutionData, Split, Splits, merge_segments
from ._fit_metrics import get_evaluators
from . import _boosting_opt as opt


def to_array(ndvar: Union[NDVar, Tuple[NDVar, ...], float]) -> np.ndarray:
    if isinstance(ndvar, NDVar):
        return ndvar.x.ravel()
    elif isinstance(ndvar, tuple):
        return np.concatenate([to_array(x) for x in ndvar])
    else:
        return np.array([ndvar])


@dataclass(eq=False)
class BoostingResult(PickleableDataClass):
    """Result from boosting

    Fit metrics are computed from all time points in ``y``.

    - For models estimated with ``test=False``, the entire ``y_pred`` is
      predicted with the averaged TRF from the different runs (with each run
      corresponding to one validation set).
    - For models estimated with ``test=True``, ``y_pred`` in each test segment
      is predicted from the averaged TRF from the corresponding training runs
      (each non-test segment used as validation set once), and the different
      test segments are then concatenated to compute the fit-metrics in
      comparison with ``y``.


    Attributes
    ----------
    h : NDVar | tuple of NDVar
        The temporal response function (the average of all TRFs from the
        different runs/partitions).
        Whether ``h`` is an :class:`NDVar` or a :class:`tuple` of :class:`NDVar`
        depends on whether the ``x`` parameter to :func:`boosting` was an
        :class:`NDVar` or a sequence of :class:`NDVar`.
    h_scaled : NDVar | tuple of NDVar
        ``h`` scaled such that it applies to the original input ``y`` and ``x``.
        If boosting was done with ``scale_data=False``, ``h_scaled`` is the same
        as ``h``.
    h_source : NDVar | tuple of NDVar
        If ``h`` was constructed using a basis, ``h_source`` represents the
        source of ``h`` before being convolved with the basis.
    h_time : UTS
        Time dimension of the kernel.
    r : float | NDVar
        Correlation between the measured response ``y`` and the predicted
        response ``h * x``. When using cross-validation (calling
        :func:`boosting` with ``test=True``), each partition of ``y`` is
        predicted using the ``h`` estimated from the corresponding training
        partitions. Otherwise, all of ``y`` is estimated using the average ``h``.
        For vector data, measured and predicted responses are normalized, and ``r``
        is computed as the average dot product over time.
        The type of ``r`` depends on the ``y`` parameter to :func:`boosting`:
        If ``y`` is one-dimensional, ``r`` is scalar, otherwise it is a :class:`NDVar`.
        Note that ``r`` does not take into account the model's ability to predict
        the magnitude of the response, only its shape; for a measure that reflects both,
        consider using ``proportion_explained``.
    r_rank : float | NDVar
        As ``r``, the Spearman rank correlation.
    t_run : float
        Time it took to run the boosting algorithm (in seconds).
    error : str
        The error evaluation method used.
    residual : float | NDVar
        The residual of the final result

        - ``error='l1'``: the sum of the absolute differences between ``y`` and
          ``h * x``.
        - ``error='l2'``: the sum of the squared differences between ``y`` and
          ``h * x``.

        For vector ``y``, the error is defined based on the distance in space
        for each data point.
    delta : scalar
        Kernel modification step used.
    mindelta : None | scalar
        Mindelta parameter used.
    n_samples : int
        Number of samples in the input data time axis.
    proportion_explained : float | NDVar
        The proportion of the variation in ``y`` that is explained by the model.
        Calculated as ``1 - (variation(residual) / variation(y))``.
        Variation is caculated as the ``l1`` or ``l2`` norm,
        depending on the ``error`` that was used for model fitting.
        Note that this does not correspond to ``r**2`` even for ``error='l2'``,
        because residuals are not guaranteed to be orthogonal to predictions.
    scale_data : bool
        Scale_data parameter used.
    y_mean : NDVar | scalar
        Mean that was subtracted from ``y``.
    y_scale : NDVar | scalar
        Scale by which ``y`` was divided.
    x_mean : NDVar | scalar | tuple
        Mean that was subtracted from ``x``.
    x_scale : NDVar | scalar | tuple
        Scale by which ``x`` was divided.
    splits : Splits
        Data splits used for cross-validation.
        Use :meth:`.splits.plot` to visualize the cross-validation scheme.
    partition_results : list of BoostingResuls
        If :func:`boosting` is called with ``partition_results=True``, this
        attribute contains the results for the individual test paritions.
    algorithm_version : int
        Version of the algorithm with which the model was estimated

          - -1: results from before this attribute was added
          - 0: Normalize ``x`` after applying basis
          - 1: Numba implementation
          - 2: Cython multiprocessing implementation (Eelbrain 0.38)
    """
    # basic parameters
    y: Optional[str]
    x: Union[Optional[str], Tuple[Optional[str]]]
    tstart: Union[float, Tuple[float, ...]]
    tstop: Union[float, Tuple[float, ...]]
    scale_data: bool
    delta: float
    mindelta: float
    error: str
    selective_stopping: int
    # data properties
    y_mean: NDVar
    y_scale: NDVar
    x_mean: Union[NDVar, Tuple[NDVar, ...]]
    x_scale: Union[NDVar, Tuple[NDVar, ...]]
    # results
    _h: Union[NDVar, Tuple[NDVar, ...]]
    _isnan: np.ndarray
    t_run: float
    # advanced parameters
    basis: float
    basis_window: str
    splits: Splits = None
    # advanced data properties
    n_samples: int = None
    _y_info: dict = field(default_factory=dict)
    _y_dims: Tuple[Dimension, ...] = None
    # fit metrics
    i_test: int = None  # test partition for fit metrics
    residual: Union[float, NDVar] = None
    r: Union[float, NDVar] = None
    r_rank: Union[float, NDVar] = None
    r_l1: NDVar = None
    partition_results: List[BoostingResult] = None
    # store the version of the boosting algorithm with which model was fit
    version: int = 13  # file format (updates when re-saving)
    algorithm_version: int = -1  # does not change when re'saving
    # debug parameters
    y_pred: NDVar = None
    fit: Any = None  # scanpydoc can't handle undocumented 'Boosting'
    # Legacy attributes (pickled objects may have these attributes)
    prefit: str = None

    def __post_init__(self):
        if self.splits is None:
            self.partitions = None
        else:
            self.partitions = self.splits.n_partitions

    def __setstate__(self, state: dict):
        # backwards compatibility
        version = state.pop('version')
        if version < self.version:
            if version == 7:
                state['partitions'] = state.pop('n_partitions')
                state['partitions_arg'] = state.pop('n_partitions_arg')
            if version < 9:
                state['residual'] = state.pop('fit_error')
            if version < 11:
                for key in ['partitions_arg', 'h', 'isnan', 'y_info']:
                    state[f'_{key}'] = state.pop(key, None)
                state['r_rank'] = state.pop('spearmanr')
            if version < 12:
                # Vector residuals are averaged
                if state['r_rank'] is None:
                    state['residual'] *= state['n_samples']
            if version < 13:
                state['splits'] = Splits(None, state.pop('_partitions_arg'), state.pop('partitions'), state.pop('validate', 1), state.pop('test', 0), state.pop('model'))
        PickleableDataClass.__setstate__(self, state)

    def __repr__(self):
        items = []
        if isinstance(self.tstart, tuple):
            x = ' + '.join(f'{x} ({t1:g} - {t2:g})' for x, t1, t2 in zip(self.x, self.tstart, self.tstop))
        else:
            if self.x is None or isinstance(self.x, str):
                x = self.x
            else:
                x = ' + '.join(map(str, self.x))
            items.append(f'{self.tstart:g} - {self.tstop:g}')
        items.insert(0, f'boosting {self.y} ~ {x}')
        for name, param in inspect.signature(boosting).parameters.items():
            if param.default is inspect.Signature.empty or name == 'data':
                continue
            elif name == 'debug':
                continue
            elif name == 'partition_results':
                value = bool(self.partition_results)
            elif name == 'partitions':
                value = None if self.splits is None else self.splits.partitions_arg
            elif name == 'model':
                if self.splits is None or self.splits.model is None:
                    continue
                value = dataobj_repr(self.splits.model)
            elif name == 'validate':
                value = None if self.splits is None else self.splits.n_validate
            elif name == 'test':
                value = None if self.splits is None else self.splits.n_test
            else:
                value = getattr(self, name)
            if value is not None and value != param.default:
                items.append(f'{name}={value}')
        return f"<{', '.join(items)}>"

    @cached_property
    def _x_mean_array(self):
        return to_array(self.x_mean)

    @cached_property
    def _x_scale_array(self):
        return to_array(self.x_scale)

    @cached_property
    def h(self):
        if not self.basis:
            return self._h
        elif isinstance(self._h, tuple):
            return tuple(h.smooth('time', self.basis, self.basis_window, 'full') for h in self._h)
        else:
            return self._h.smooth('time', self.basis, self.basis_window, 'full')

    @cached_property
    def h_scaled(self):
        if self.y_scale is None:
            return self.h
        elif isinstance(self.h, NDVar):
            out = self.h * (self.y_scale / self.x_scale)
            out.info = self._y_info.copy()
            return out
        else:
            out = []
            for h, sx in zip(self.h, self.x_scale):
                h = h * (self.y_scale / sx)
                h.info = self._y_info.copy()
                out.append(h)
            return tuple(out)

    @cached_property
    def h_source(self):
        return self._h

    @cached_property
    def h_time(self):
        if isinstance(self.h, NDVar):
            return self.h.time
        else:
            return self.h[0].time

    @cached_property
    def _variability(self):
        # variability in the data
        if self.y_scale is None:
            raise NotImplementedError("Not implemented for scale_data=False")
        elif self.n_samples is None:
            raise OldVersionError("This is an older result object which did not store some necessary information; refit the model to use this attribute")
        else:
            # Due to the normalization:
            return self.n_samples

    @deprecate_ds_arg
    def cross_predict(
            self,
            x: Union[NDVarArg, Sequence[NDVarArg]] = None,
            data: Dataset = None,
            scale: Literal['original', 'normalized'] = 'original',
            name: str = None,
    ) -> NDVar:
        """Predict responses to ``x`` using complementary training data

        Parameters
        ----------
        x
            Predictors used in the original model fit, or a subset thereof.
            In order for cross-prediction to be accurate, ``x`` needs to match the ``x``
            used in the original fit exactly in cases and time.
        data
            Dataset with predictors. If ``ds`` is specified, ``x`` can be omitted.
        scale
            Return predictions at the scale of the original data (the ``y``
            supplied to the :func:`boosting` function) or at the normalized
            scale that is used for model fitting (``y`` and ``x`` normalized).
        name
            Name for the output :class:`NDVar`.

        See Also
        --------
        convolve : Simple prediction of linear model

        Notes
        -----
        This function does not adjust the mean across time of predicted
        responses; subtract the mean in order to compute explained variance.

        Examples
        --------
        Fit a TRF and reproduce the error using the cross-predict function::

            trf = boosting(y, x, 0, 0.5, partitions=5, test=1)
            y_pred = trf.cross_predict(x, scale='normalized')

            y_normalized = (y - trf.y_mean) / trf.y_scale
            y_residual = y_normalized - y_pred
            proportion_explained_l1 = 1 - (y_residual.abs().sum('time') / y_normalized.abs().sum('time'))
            proportion_explained_l2 = 1 - ((y_residual ** 2).sum('time') / (y_normalized ** 2).sum('time'))

        """
        if scale not in ('original', 'normalized'):
            raise ValueError(f"{scale=}")
        if not self.partition_results:
            raise ValueError("BoostingResult does not contain partition-specific models; fit with partition_results=True")
        # predictors
        x_ = self.x if x is None else x
        x_data = PredictorData(x_, data, copy=True)
        # check predictors match h
        if x_data.x_name == self.x:
            x_use = None
        else:
            if isinstance(self.x, str):
                raise ValueError(f'{x=} for {self}')
            elif x_data.multiple_x:
                x_use = x_data.x_name
            else:
                x_use = [x_data.x_name]
            missing = set(x_use).difference(self.x)
            if missing:
                raise ValueError(f"{x=}: has variables not in {self}:\nx: {', '.join(missing)}")
        # prepare output array
        if self._y_dims is None:  # only possible in results from dev version
            y_dims = self.y_mean.dims
        else:
            y_dims = self._y_dims
        y_dimnames = [dim.name for dim in y_dims]
        n_y = sum(len(dim) for dim in y_dims) or 1
        y_pred = np.empty((1, n_y, x_data.n_times_flat))
        # prepare x:  (n_x_only, n_shared, n_x_times)
        x_array = x_data.data
        if self.scale_data:
            x_array -= self._x_mean_array[:, np.newaxis]
            x_pads = -(self._x_mean_array / self._x_scale_array)[np.newaxis]
            x_array /= self._x_scale_array[:, np.newaxis]
        else:
            x_pads = np.zeros((1, len(x_array)))
        x_array = x_array[np.newaxis, :, :]
        # prepare h
        h_i_start = int(round(self.h_time.tmin / self.h_time.tstep))
        # iterate through partitions
        for result in self.partition_results:
            # find segments
            for split in self.splits.splits:
                if split.i_test == result.i_test:
                    segments = split.test
                    break
            else:
                raise RuntimeError(f"Split missing for test segment {result.i_test}")
            # h to flat array: (n_h_only == in y, n_shared == in x, n_h_times)
            hs = [result.h] if isinstance(result.h, NDVar) else result.h
            if x_use:
                hs = {h.name: h for h in hs}
                hs = [hs[x] for x in x_use]
            h_array = []
            for h, (name, xdims, index) in zip(hs, x_data.x_meta):
                dimnames = [*y_dimnames, *[dim.name for dim in xdims], 'time']
                h_data = h.get_data(dimnames)
                h_data = h_data.reshape((n_y, -1, h_data.shape[-1]))
                h_array.append(h_data)
            h_array = np.concatenate(h_array, 1)
            # convolution
            parallel_convolve_segments(h_array, x_array, x_pads, h_i_start, segments, y_pred)
        # package output
        dims = [*y_dims, x_data.time_dim]
        shape = [len(dim) for dim in dims]
        if x_data.case_to_segments:
            dims = (Case, *dims)
            shape.insert(-1, x_data.n_cases)
        y_pred = y_pred.reshape(shape)
        if x_data.case_to_segments:
            y_pred = np.rollaxis(y_pred, -2)
        if name is None:
            name = self.y
        y_pred = NDVar(y_pred, dims, name, self._y_info)
        if scale == 'original' and self.scale_data:
            y_pred *= self.y_scale
            y_pred += self.y_mean
        return y_pred

    def partition_result_data(self, model: str = None) -> Dataset:
        """Results from the different test partitions in a :class:`Dataset`

        Parameters
        ----------
        model
            Add a ``'model'`` column to the dataset to distinguish
        """
        h_is_list = isinstance(self._h, tuple)
        rows = []
        for res in self.partition_results:
            hs = res.h if h_is_list else [res.h]
            rows.append([res.i_test, res.r, res.proportion_explained, *hs])
        if self.x in (None, (None,)):
            xs = ['x']
        elif isinstance(self.x, str):
            xs = [self.x]
        else:
            xs = [f'x_{i}' if x is None else x for i, x in enumerate(self.x)]
        return Dataset.from_caselist(['i_test', 'r', 'det', *xs], rows)

    @cached_property
    def proportion_explained(self):
        return 1 - (self.residual / self._variability)

    def _apply_ndvar_transform(self, func: Callable):
        "Apply func to all NDVars in-place (only for source space transformation)"
        def sub_func(obj):
            if obj is None:
                return None
            elif isinstance(obj, tuple):
                return tuple(sub_func(obj_) for obj_ in obj)
            return func(obj)

        # NDVars
        for attr in ('_h', 'r', 'r_rank', 'residual', 'y_mean', 'y_scale'):
            setattr(self, attr, sub_func(getattr(self, attr)))

        # List of Dimension
        if self._y_dims is not None:
            self._y_dims = tuple([sub_func(dim) if isinstance(dim, SourceSpaceBase) else dim for dim in self._y_dims])

        if self.partition_results:
            for res in self.partition_results:
                res._apply_ndvar_transform(func)

    def _morph(self, to_subject: str):
        "Morph source space"
        from .._mne import morph_source_space

        def func(obj: NDVar):
            return morph_source_space(obj, to_subject)
        self._apply_ndvar_transform(func)

    def _set_connectivity(self, dim, connectivity):
        def func(obj: NDVar):
            return set_connectivity(obj, dim, connectivity)
        self._apply_ndvar_transform(func)

    def _set_parc(self, parc: str):
        """Change the parcellation of source-space result

        Notes
        -----
        No warning for missing sources!
        """
        def func(obj: NDVar):
            return set_parc(obj, parc, mask=True)
        self._apply_ndvar_transform(func)

    @classmethod
    def _eelbrain_concatenate(
            cls,
            results: Sequence[BoostingResult],
            dim: str = 'source',
    ):
        "Combine multiple complementary source-space BoostingResult objects"
        # result = results[0]
        out = {}
        for field in fields(cls):
            if field.name == 'version':
                continue
            elif field.name == '_isnan':
                out['_isnan'] = None
                continue
            values = [getattr(result, field.name) for result in results]
            if field.name == 't_run':
                out['t_run'] = None if any(v is None for v in values) else sum(values)
                continue
            if field.name == 'partition_results' and any(v is not None for v in values):
                if not all(v is not None for v in values):
                    raise ValueError(f'partition_results avaiable for some but not all part-results')
                new_value = [cls._eelbrain_concatenate(p_results) for p_results in zip(*values)]
            elif field.name == 'algorithm_version':
                values = set(values)
                if len(values) == 1:
                    new_value = values.pop()
                else:
                    new_value = tuple(sorted(values))
            elif any(v is None for v in values):
                new_value = None
            else:
                new_value = _concatenate_values(values, dim, field.name)
            out[field.name] = new_value
        return cls(**out)


class SplitResult:
    __slots__ = ('split', 'h', 'h_failed')

    def __init__(self, split: Split, h: np.ndarray, h_failed: np.ndarray):
        self.split = split
        self.h = h  # (n_y, n_x, n_times_h)
        self.h_failed = h_failed  # (n_y,)

    def h_with_nan(self):
        "Set failed TRFs to NaN"
        if np.any(self.h_failed):
            h = self.h.copy()
            h[self.h_failed] = np.nan
            return h
        return self.h


class Boosting:
    """Object-oriented API for boosting

    Examples
    --------
    Standard usage of the object-oriented API for a model with
    cross-validation, comparable to the :func:`boosting` function::

        data = DeconvolutionData(y, x, ds)
        data.apply_basis(0.05, 'hamming'')
        data.normalize('l1')
        data.initialize_cross_validation(5, test=1)
        model = Boosting(data)
        model.fit(0, 0.500, selective_stopping=1, error='l1')
        result = model.evaluate_fit()
    """
    # fit parameters
    tstart = None
    tstart_h = None
    tstop = None
    selective_stopping = None
    error = None
    delta = None
    mindelta = None
    # timing
    t_fit_start = None
    t_fit_done = None
    # fit result
    _i_start = None
    split_results = None
    n_skip = 0
    # eval result
    y_pred = None

    def __init__(
            self,
            data: DeconvolutionData,
    ):
        self.data = data

    def fit(
            self,
            tstart: Union[float, Sequence[float]],
            tstop: Union[float, Sequence[float]],
            selective_stopping: int = 1,
            error: str = 'l1',
            delta: float = 0.005,  # coordinate search step
            mindelta: float = None,  # narrow search by reducing delta until reaching mindelta
    ):
        self.data._check_data()
        assert error in ('l1', 'l2')
        error_id = int(error[1])
        mindelta_ = delta if mindelta is None else mindelta
        self.selective_stopping = selective_stopping
        self.error = error
        self.delta = delta
        self.mindelta = mindelta
        n_x = len(self.data.x)
        # find TRF start/stop for each x
        if isinstance(tstart, (tuple, list, np.ndarray)):
            if len(tstart) != len(self.data.x_name):
                raise ValueError(f'{tstart=}: {len(tstart)=} is different from len(x)={len(self.data.x_name)}')
            elif len(tstart) != len(tstop):
                raise ValueError(f'{tstop=}: {len(tstop)=} does not match {len(tstart)=}')
            self.tstart = tuple(tstart)
            self.tstart_h = min(self.tstart)
            self.tstop = tuple(tstop)
            if any(start >= stop for start, stop in zip(self.tstart, self.tstop)):
                raise ValueError(f"Some tstart > tstop: {tstart=} and {tstop=}")
            n_xs = [reduce(mul, map(len, xdims), 1) for _, xdims, _ in self.data._x_meta]
            tstart = [t for t, n in zip(tstart, n_xs) for _ in range(n)]
            tstop = [t for t, n in zip(tstop, n_xs) for _ in range(n)]
        else:
            if tstart >= tstop:
                raise ValueError(f"{tstart=} > {tstop=}")
            self.tstart = self.tstart_h = tstart
            self.tstop = tstop
            tstart = [tstart] * n_x
            tstop = [tstop] * n_x

        # TRF extent in indices
        tstep = self.data.time.tstep
        i_start_by_x = np.asarray([int(round(t / tstep)) for t in tstart], np.int64)
        i_stop_by_x = np.asarray([int(ceil(t / tstep)) for t in tstop], np.int64)
        self._i_start = i_start = np.min(i_start_by_x)
        i_stop = np.max(i_stop_by_x)
        h_n_times = i_stop - i_start

        if len(self.data.segments) == 1:
            self.n_skip = h_n_times - 1

        self.t_fit_start = time.time()

        # boosting
        num_threads = CONFIG['n_workers']
        split_train = package_splits([split.train for split in self.data.splits.splits])
        split_validate = package_splits([split.validate for split in self.data.splits.splits])
        split_train_and_validate = package_splits([split.train_and_validate for split in self.data.splits.splits])
        hs, hs_failed = opt.boosting_runs(self.data.y, self.data.x, self.data.x_pads, split_train, split_validate, split_train_and_validate, i_start_by_x, i_stop_by_x, delta, mindelta_, error_id, selective_stopping, num_threads)
        self.split_results = [SplitResult(split, h, h_failed) for split, h, h_failed in zip(self.data.splits.splits, hs, hs_failed)]

        self.t_fit_done = time.time()

    def _get_i_tests(self):
        assert self.data.splits.n_test
        return sorted({split.split.i_test for split in self.split_results})

    def _get_h(
            self,
            skip_failed: bool = False,
            i_test: int = None,  # test partition
    ):  # kernel and sgements on which to evaluate
        if i_test is None:
            split_results = self.split_results
            segments = self.data.segments
        else:
            split_results = [split for split in self.split_results if split.split.i_test == i_test]
            segments = split_results[0].split.test

        if skip_failed:
            with warnings.catch_warnings():
                warnings.filterwarnings('ignore', 'Mean of empty slice', RuntimeWarning)
                h = np.nanmean([split.h_with_nan() for split in split_results], 0)
            is_nan = np.isnan(h[:, :, 0])
            h[is_nan] = 0
        else:
            h = np.mean([split.h for split in split_results])
        return h, segments

    def _get_h_failed(self, i_test: int = None) -> np.ndarray:
        if i_test is None:
            split_results = self.split_results
        else:
            split_results = [split for split in self.split_results if split.split.i_test == i_test]
        out = np.all([split.h_failed for split in split_results], 0)
        if self.data.vector_dim:
            out = np.all(out.reshape((len(self.data.vector_dim), -1)), 0)
        return out

    def evaluate_fit(
            self,
            i_test: int = None,
            metrics: Sequence[str] = None,
            cross_fit: bool = None,
            partition_results: bool = False,
            debug: bool = False,
    ):
        """Compute average TRF and fit metrics

        Parameters
        ----------
        i_test
            Test partition index (only applies for models fit with
            cross-validation). By default, ``y`` in each test segment is
            predicted from the averaged model from the corresponding training
            sets, and all test segments are pooled for computing fit metrics.
            Set ``i_test`` with an integer to compute fit metrics for a single
            test partition.
        metrics
            Which model fit metrics to compute (default depends on the data).
            If no metrics are needed, ``metrics=()`` is faster.
        cross_fit
            Compute fit metrics from cross-validation (only applies to model
            with cross-validation; default ``True``).
        partition_results
            Keep results (TRFs and model evaluation) for each test-partition.
        debug
            Add additional attributes to the returned result.
        """
        if cross_fit is None:
            cross_fit = bool(self.data.splits.n_test)
        elif cross_fit and not self.data.splits.n_test:
            raise ValueError(f"{cross_fit=} for model without cross-validation")
        if partition_results and not cross_fit:
            raise ValueError(f"{partition_results=} with {cross_fit=}")

        # fit evaluation
        if metrics is None:
            if self.data.vector_dim:
                metrics = [f'vec-{self.error}', f'vec-corr']
                if self.error == 'l1':
                    metrics.append('vec-corr-l1')
            else:
                metrics = [self.error, 'r', 'r_rank']

        # test sets to use
        if cross_fit:
            if i_test is None:
                i_tests = self._get_i_tests()
            else:
                i_tests = [i_test]
        elif i_test is not None:
            raise ValueError(f"{i_test=} without cross_fit")
        else:
            i_tests = None

        # hs: [(h, test_segments), ...]
        if cross_fit:
            hs = [self._get_h(True, i) for i in i_tests]
            all_segments = np.sort(np.vstack([segments for _, segments in hs]), 0)
            eval_segments = [merge_segments(all_segments, True)]
            if partition_results:
                for _, test_segments in hs:
                    eval_segments.append(merge_segments(test_segments, True))
        else:
            hs = [self._get_h(True)]
            eval_segments = merge_segments(self.data.segments, True)
            if self.n_skip:
                eval_segments[:, 0] += self.n_skip
                # check for invalid segments (negative duration)
                valid = eval_segments[:, 1] - eval_segments[:, 0] > 0
                if not np.all(valid):
                    eval_segments = eval_segments[valid]
            eval_segments = [eval_segments]

        if metrics:
            # y dimensions
            n_y = len(self.data.y)
            if self.data.vector_dim:
                n_vec = len(self.data.vector_dim)
                n_vecs = n_y // n_vec
            else:
                n_vec = n_vecs = 0

            # predicted y
            if debug:
                y_pred_iter = y_pred = np.empty(self.data.y.shape)
            elif n_vecs:
                y_pred = np.empty((n_vec, *self.data.y.shape[1:]))
                y_pred_iter = chain.from_iterable(repeat(tuple(y_pred), n_vecs))
            else:
                y_pred = np.empty(self.data.y.shape[1:])
                y_pred_iter = repeat(y_pred, n_y)

            all_evaluators, evaluators_s, evaluators_v = get_evaluators(metrics, self.data, eval_segments)

            # fit and evaluate each y
            for i_y, y_pred_i in enumerate(y_pred_iter):
                # for cross-validation, different segments are predicted by different h:
                for h, segments in hs:
                    convolve_segments_jit(h[i_y], self.data.x, self.data.x_pads, self._i_start, segments, y_pred_i)

                if evaluators_s:
                    for e in evaluators_s:
                        e.add_y(i_y, self.data.y[i_y], y_pred_i)

                if evaluators_v and i_y % n_vec == n_vec - 1:
                    i_vec = i_y // n_vec
                    i_y_vec = slice(i_y-n_vec+1, i_y+1)
                    y_pred_i_vec = y_pred[i_y_vec] if debug else y_pred
                    for e in evaluators_v:
                        e.add_y(i_vec, self.data.y[i_y_vec], y_pred_i_vec)

            # Package evaluators
            evaluations = {e.attr: e.get() for e in all_evaluators}
            if debug:
                evaluations['y_pred'] = self.data.package_y_like(y_pred, 'y-pred')
            if partition_results:
                partition_evaluations = {i: {e.attr: e.get(i) for e in all_evaluators} for i in i_tests}
            else:
                partition_evaluations = None
        else:
            evaluations = {}
            partition_evaluations = {i: {} for i in i_tests}

        # package h
        h_xs = [h for h, _ in hs]
        h_x = h_xs[0] if len(h_xs) == 1 else np.mean(h_xs, 0)
        h = self.data.package_kernel(h_x, self.tstart_h)
        # package model parameters
        y_mean, y_scale, x_mean, x_scale = self.data.data_scale_ndvars()
        if debug:
            evaluations['fit'] = self
        t_run = self.t_fit_done - self.t_fit_start
        # partition-specific results
        if partition_results:
            partition_results_list = []
            for i in i_tests:
                h_i = self.data.package_kernel(h_xs[i], self.tstart_h)
                evaluations_i = partition_evaluations[i]
                result = BoostingResult(
                    self.data.y_name, self.data.x_name, self.tstart, self.tstop, bool(self.data.scale_data), self.delta, self.mindelta, self.error, self.selective_stopping,
                    y_mean, y_scale, x_mean, x_scale,
                    h_i, self._get_h_failed(i), 0,
                    self.data.basis, self.data.basis_window, None,
                    self.data.y.shape[1], self.data.y_info, self.data.ydims,
                    algorithm_version=2, i_test=i, **evaluations_i)
                partition_results_list.append(result)
        else:
            partition_results_list = None
        return BoostingResult(
            # basic parameters
            self.data.y_name, self.data.x_name, self.tstart, self.tstop, bool(self.data.scale_data), self.delta, self.mindelta, self.error, self.selective_stopping,
            # data properties
            y_mean, y_scale, x_mean, x_scale,
            # results
            h, self._get_h_failed(), t_run,
            # advanced parameters
            self.data.basis, self.data.basis_window, self.data.splits,
            # advanced data properties
            self.data.y.shape[1], self.data.y_info, self.data.ydims,
            partition_results=partition_results_list,
            algorithm_version=2,
            i_test=i_test, **evaluations)


@user_activity
@deprecate_ds_arg
def boosting(
        y: NDVarArg,
        x: Union[NDVarArg, Sequence[NDVarArg]],
        tstart: Union[float, Sequence[float]],
        tstop: Union[float, Sequence[float]],
        scale_data: Union[bool, str] = True,  # normalize y and x; can be 'inplace'
        delta: float = 0.005,  # coordinate search step
        mindelta: float = None,  # narrow search by reducing delta until reaching mindelta
        error: Literal['l1', 'l2'] = 'l2',
        basis: float = 0,
        basis_window: str = 'hamming',
        partitions: int = None,  # Number of partitionings for cross-validation
        model: CategorialArg = None,
        validate: int = 1,  # Number of segments in validation set
        test: int = 0,  # Number of segments in test set
        data: Dataset = None,
        selective_stopping: int = 0,
        partition_results: bool = False,
        debug: bool = False,
) -> BoostingResult:
    """Estimate a linear filter with coordinate descent

    Parameters
    ----------
    y : NDVar
        Signal to predict. When ``y`` contains more than one signal (e.g.,
        multiple EEG channels), results for each signal will be computed
        independently. Muiltiple cases along a :class:`Case` dimension are
        treated as different trials which share a filter. For correlation fit
        metrics, a :class:`Space` dimension is interpreted as defining a vector
        measure.
    x : NDVar | sequence of NDVar
        Signal to use to predict ``y``. Can be sequence of NDVars to include
        multiple predictors. Time dimension must correspond to ``y``.
    tstart : scalar | sequence of scalar
        Start of the TRF in seconds. A list can be used to specify different
        values for each item in ``x``.
    tstop : scalar | sequence of scalar
        Stop of the TRF in seconds. Format must match ``tstart``.
    scale_data : bool | 'inplace'
        Scale ``y`` and ``x`` before boosting: subtract the mean and divide by
        the standard deviation (when ``error='l2'``) or the mean absolute
        value (when ``error='l1'``). Use ``'inplace'`` to save memory by scaling
        the original objects specified as ``y`` and ``x`` instead of making a 
        copy. The data scale is stored in the :class:`BoostingResult:
        :attr:`.y_mean``, :attr:`.y_scale`, :attr:`.x_mean`, and :attr:`.x_scale`
        attributes.
    delta
        Step for changes in the kernel.
    mindelta
        If the error for the training data can't be reduced, divide ``delta``
        in half until ``delta < mindelta``. The default is ``mindelta = delta``,
        i.e. ``delta`` is constant.
    error
        Error function to use (default is ``l2``).

        - ``error='l1'``: the sum of the absolute differences between ``y`` and
          ``h * x``.
        - ``error='l2'``: the sum of the squared differences between ``y`` and
          ``h * x``.

        For vector ``y``, the error is defined based on the distance in space
        for each data point.
    basis
        Use a basis of windows with this length for the kernel (by default,
        impulses are used).
    basis_window : str | scalar | tuple
        Basis window (see :func:`scipy.signal.get_window` for options; default
        is ``'hamming'``).
    partitions
        Divide the data into this many ``partitions`` for cross-validation-based
        early stopping. In each partition, ``n - 1`` segments are used for
        training, and the remaining segment is used for validation.
        If data is continuous, data are divided into contiguous segments of
        equal length (default 10).
        If data has cases, cases are divided with ``[::partitions]`` slices
        (default ``min(n_cases, 10)``; if ``model`` is specified, ``n_cases``
        is the lowest number of cases in any cell of the model).
        See :ref:`exa-data_split` example.
    model
        If data has cases, divide cases into different categories (division
        for crossvalidation is done separately for each cell).
    data
        If provided, other parameters can be specified as string for items in
        ``ds``.
    validate
        Number of segments in validation dataset (currently has to be 1).
    test
        By default (``test=0``), the boosting algorithm uses all available data
        to estimate the kernel. Set ``test=1`` to perform *k*-fold cross-
        validation instead (with *k* = ``partitions``):
        Each partition is used as test dataset in turn, while the remaining
        ``k-1`` partitions are used to estimate the kernel. The resulting model
        fit metrics reflect the re-combination of all partitions, each one
        predicted from the corresponding, independent training set.
    selective_stopping
        By default, the boosting algorithm stops when the testing error stops
        decreasing. With ``selective_stopping=True``, boosting continues but
        excludes the predictor (one time-series in ``x``) that caused the
        increase in testing error, and continues until all predictors are
        stopped. The integer value of ``selective_stopping`` determines after
        how many steps with error increases each predictor is excluded.
    partition_results
        Keep results (TRFs and model evaluation) for each test-partition.
        This is disabled by default to reduce file size when saving results.
    debug
        Add additional attributes to the returned result.

    See Also
    --------
    plot.preview_partitions : preview data partitions for cross-validation

    Notes
    -----
    The boosting algorithm is described in [1]_.

    In order to predict data, use the :func:`convolve` function::

    >>> ds = datasets.get_uts()
    >>> data['a1'] = epoch_impulse_predictor('uts', 'A=="a1"', ds=data)
    >>> data['a0'] = epoch_impulse_predictor('uts', 'A=="a0"', ds=data)
    >>> res = boosting('uts', ['a0', 'a1'], 0, 0.5, partitions=10, model='A', data=data)
    >>> y_pred = convolve(res.h_scaled, ['a0', 'a1'], ds=data)
    >>> y = data['uts']
    >>> plot.UTS([y-y.mean('time'), y_pred], '.case')

    References
    ----------
    .. [1] David, S. V., Mesgarani, N., & Shamma, S. A. (2007). Estimating
        sparse spectro-temporal receptive fields with natural stimuli. Network:
        Computation in Neural Systems, 18(3), 191-212.
        `10.1080/09548980701609235 <https://doi.org/10.1080/09548980701609235>`_.
    """
    # scale_data
    if isinstance(scale_data, bool):
        scale_in_place = False
    elif isinstance(scale_data, str):
        if scale_data == 'inplace':
            scale_in_place = True
        else:
            raise ValueError(f"{scale_data=}")
    else:
        raise TypeError(f"{scale_data=}, need bool or str")
    # selective_stopping
    selective_stopping = int(selective_stopping)
    if selective_stopping < 0:
        raise ValueError(f"{selective_stopping=}")

    dec_data = DeconvolutionData(y, x, data, scale_in_place)
    dec_data.apply_basis(basis, basis_window)
    if scale_data:
        dec_data.normalize(error)
    dec_data.initialize_cross_validation(partitions, model, data, validate, test)

    fit = Boosting(dec_data)
    fit.fit(tstart, tstop, selective_stopping, error, delta, mindelta)
    return fit.evaluate_fit(debug=debug, partition_results=partition_results)


def convolve(
        h: np.ndarray,  # (n_stims, h_n_samples)
        x: np.ndarray,  # (n_stims, n_samples)
        x_pads: np.ndarray,  # (n_stims,)
        h_i_start: int,
        segments: np.ndarray = None,  # (n_segments, 2)
        out: np.ndarray = None,
) -> np.ndarray:
    """h * x with time axis matching x

    Parameters
    ----------
    h
        H.
    x
        X.
    x_pads
        Padding for x.
    h_i_start
        Time shift of the first sample of ``h``.
    segments
        Data segments.
    out
        Buffer for predicted ``y``.
    """
    n_x, n_times = x.shape
    if segments is None:
        segments = np.array(((0, n_times),), np.int64)
    if out is None:
        out = np.empty(n_times)
    return convolve_segments_jit(h, x, x_pads, h_i_start, segments, out)


@numba.njit(nogil=True, cache=True, parallel=True)
def parallel_convolve_segments(
        h_flat: np.ndarray,  # (n_h_only, n_shared, n_h_times)
        x_flat: np.ndarray,  # (n_x_only, n_shared, n_x_times)
        x_pads: np.ndarray,  # (n_x_only, n_shared)
        h_i_start: int,  # offset of the first sample of h
        segments: np.ndarray,  # (n_segments, 2)
        out_flat: np.ndarray,  # (n_x_only, n_h_only, n_x_times)
):
    # loop through x and h dimensions
    out_indexes = [(ix, ih) for ix in range(len(x_flat)) for ih in range(len(h_flat))]
    for i_out in numba.prange(len(out_indexes)):
        i_x, i_h = out_indexes[i_out]
        convolve_segments_jit(h_flat[i_h], x_flat[i_x], x_pads[i_x], h_i_start, segments, out_flat[i_x, i_h])


@numba.njit(nogil=True, cache=True)
def convolve_segments_jit(
        h: np.ndarray,  # (n_x, n_h_times)
        x: np.ndarray,  # (n_x, n_x_times)
        x_pads: np.ndarray,  # (n_x,)
        h_i_start: int,  # offset of the first sample of h
        segments: np.ndarray,  # (n_segments, 2)
        out: np.ndarray,  # (n_x_times,)
):
    h_n_times = h.shape[1]
    segments_start = segments[:, 0]
    segments_stop = segments[:, 1]
    for start, stop in zip(segments_start, segments_stop):
        out[start: stop] = 0
    h_i_stop = h_i_start + h_n_times

    # padding
    h_pad = np.sum(h * x_pads.reshape((-1, 1)), 0)
    # padding for pre-
    pad_head_n_times = max(0, h_n_times + h_i_start)
    if pad_head_n_times:
        pad_head = np.zeros(pad_head_n_times)
        for i in range(min(pad_head_n_times, h_n_times)):
            pad_head[:pad_head_n_times - i] += h_pad[- i - 1]
    # padding for post-
    pad_tail_n_times = -min(0, h_i_start)
    if pad_tail_n_times:
        pad_tail = np.zeros(pad_tail_n_times)
        for i in range(pad_tail_n_times):
            pad_tail[i:] += h_pad[i]

    for start, stop in zip(segments_start, segments_stop):
        if pad_head_n_times:
            out[start: start + pad_head_n_times] += pad_head
        if pad_tail_n_times:
            out[stop - pad_tail_n_times: stop] += pad_tail
        convolve_jit(h, x[:, start:stop], out[start:stop], h_i_start, h_i_stop)
    return out


def package_splits(splits: Sequence[np.ndarray]) -> np.ndarray:
    n = max(len(split) for split in splits)
    out = np.empty((len(splits), n, 2), np.int64)
    out.fill(-1)
    for i, split in enumerate(splits):
        out[i, :len(split)] = split
    return out



"""Serialization with :mod:`pyarrow` (requires pyarrow to be installed)

Usage
-----

A context implements serialization of relevant Eelbrain objects, and can be
re-used indefinitely::

    >>> context = pyarrow_context()
    >>> s = context.serialize(ds).to_buffer()
    >>> ds = context.deserialize(s)
"""
from inspect import isclass
import os

from .._utils import ui


CONTEXT = None


def deserialize_2to3(obj, context):
    """Recursively deserialize objects serialized with Python 2
    
    Parameters
    ----------
    obj : object
        The object to decode.
    context : SerializationContext
        The relevant pyarrow serialization context.
    """
    if isinstance(obj, bytes):
        obj = obj.decode()
    elif isinstance(obj, dict):
        for k in tuple(obj.keys()):
            new_k = k.decode() if isinstance(k, bytes) else k
            obj[new_k] = deserialize_2to3(obj.pop(k), context)
        if context is not None and '_pytype_' in obj:
            obj = context._deserialize_callback(obj)
    elif isinstance(obj, tuple):
        obj = tuple(deserialize_2to3(item, context) for item in obj)
    elif isinstance(obj, list):
        obj = [deserialize_2to3(item, context) for item in obj]
    return obj


def serialize_factory(cls):
    if hasattr(cls, '__getstate__'):
        serialize = cls.__getstate__

        def deserialize(state):
            out = cls.__new__(cls)
            out.__setstate__(state)
            return out
    else:

        def serialize(inst):
            return inst.__reduce__()[1]

        def deserialize(data):
            return cls(*data)

    return serialize, deserialize


def pyarrow_context():
    """Generate :mod:`pyarrow` context implementing Eelbrain object serialization"""
    global CONTEXT

    if CONTEXT is None:
        import pyarrow
        import collections
        import eelbrain

        # find all classes
        classes = [
            (collections, 'OrderedDict'),
            (eelbrain._stats.testnd, 'NDPermutationDistribution'),
            (eelbrain._stats.testnd, '_MergedTemporalClusterDist'),
        ]
        for module in (eelbrain, eelbrain.test, eelbrain.testnd):
            classes.extend((module, attr) for attr in dir(module) if isclass(getattr(module, attr)))

        CONTEXT = pyarrow.SerializationContext()
        for module, attr in classes:
            cls = getattr(module, attr)
            name = '%s.%s' % (module.__name__, attr)
            serialize, deserialize = serialize_factory(cls)
            CONTEXT.register_type(cls, name, custom_serializer=serialize,
                                  custom_deserializer=deserialize)
    return CONTEXT


def load_arrow(file_path=None):
    """Load object serialized with :mod:`pyarrow`.

    Parameters
    ----------
    file_path : None | str
        Path to an arrow file. If None (default), a system file dialog will be
        shown. If the user cancels the file dialog, a RuntimeError is raised.
    """
    if file_path is None:
        filetypes = [("Parquet (*.arrow)", '*.arrow'), ("All files", '*')]
        file_path = ui.ask_file("Select Arrow file to load", "", filetypes)
        if file_path is False:
            raise RuntimeError("User canceled")
        else:
            print("load %r" % (file_path,))
    else:
        file_path = os.path.expanduser(file_path)
        if not os.path.exists(file_path):
            new_path = os.extsep.join((file_path, 'arrow'))
            if os.path.exists(new_path):
                file_path = new_path

    import pyarrow

    context = pyarrow_context()
    with pyarrow.OSFile(file_path) as fid:
        obj = context.deserialize(fid.read_buffer())

    if isinstance(obj, dict) and b'_pytype_' in obj:
        obj = deserialize_2to3(obj, context)

    return obj


def save_arrow(obj, dest=None):
    """Save a Python object with :mod:`pyarrow`.

    Parameters
    ----------
    obj : object
        Python object to save. Can contain Eelbrain objects, but custom classes
        are not supported.
    dest : None | str
        Path to destination where to save the  file. If no destination is
        provided, a file dialog is shown. If a destination without extension is
        provided, '.arrow' is appended.
    """
    if dest is None:
        filetypes = [("Arrow files (*.arrow)", '*.arrow')]
        dest = ui.ask_saveas("Save as arrow file", "", filetypes)
        if dest is False:
            raise RuntimeError("User canceled")
        else:
            print('dest=%r' % dest)
    else:
        dest = os.path.expanduser(dest)
        if not os.path.splitext(dest)[1]:
            dest += '.arrow'

    context = pyarrow_context()
    context.serialize(obj).write_to(dest)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""PySurfer Brain subclass to embed in Eelbrain"""
from distutils.version import LooseVersion
from functools import cached_property, partial
import os
import sys
from tempfile import mkdtemp
from typing import Any, Callable, Literal, Sequence, Union
from time import time, sleep
from warnings import warn

from matplotlib.colors import ListedColormap, Colormap, to_rgb, to_rgba
from matplotlib.image import imsave
from mne.io.constants import FIFF
import numpy as np
import scipy.ndimage

from .._data_obj import NDVar, SourceSpace, UTS, asndvar
from .._exceptions import KeysMissing
from .._text import ms
from .._types import PathArg
from .._utils import IS_OSX
from ..fmtxt import Image
from ..mne_fixes import reset_logger
from ._base import CONFIG, TimeSlicer, AxisScale, do_autorun, find_fig_cmaps, find_fig_vlims, fix_vlim_for_cmap, use_inline_backend
from ._color_luts import p_lut
from ._colors import ColorBar, ColorList
from ._styles import colors_for_oneway

# Traits-GUI related imports
# --------------------------
# - Set ETS toolkit before importing traits-GUI
# - Readthedocs does not support mayavi import, so we can't use surfer
# - if this is the first surfer import, lower screen logging level
first_import = 'surfer' not in sys.modules
INLINE_DISPLAY = False
try:
    from traits.trait_base import ETSConfig
    if use_inline_backend() or not CONFIG['eelbrain']:
        ETSConfig.toolkit = 'qt'
        INLINE_DISPLAY = True
        from mayavi import mlab
        mlab.options.offscreen = True
    else:
        ETSConfig.toolkit = 'wx'
    import surfer
except ImportError as exception:
    from . import _mock_surfer as surfer
    warn(f"Error importing PySurfer: {exception}")
    SURFER_IMPORTED = False
else:
    SURFER_IMPORTED = True
    if first_import:
        reset_logger(surfer.utils.logger)
del first_import


HEMI_ID_TO_STR = {FIFF.FIFFV_MNE_SURF_LEFT_HEMI: 'lh',
                  FIFF.FIFFV_MNE_SURF_RIGHT_HEMI: 'rh'}
OTHER_HEMI = {'lh': 'rh', 'rh': 'lh'}
# default size
BRAIN_H = 250
BRAIN_W = 300


def assert_can_save_movies():
    if LooseVersion(surfer.__version__) < LooseVersion('0.6'):
        raise ImportError("Saving movies requires PySurfer 0.6")


def get_source_dim(ndvar):
    if isinstance(ndvar, SourceSpace):
        return ndvar
    elif isinstance(ndvar, NDVar):
        for source in ndvar.dims:
            if isinstance(source, SourceSpace):
                return source
        raise ValueError("ndvar has no SourceSpace dimension")
    else:
        raise TypeError("ndvar=%r; needs to be NDVar or SourceSpace")


class Brain(TimeSlicer, surfer.Brain):
    """PySurfer :class:`Brain` subclass returned by :mod:`plot.brain` functions

    PySurfer :class:`surfer.Brain` subclass adding Eelbrain GUI integration and
    methods to visualize data in :class:`NDVar` format.

    Parameters
    ----------
    subject : str
        Subject name.
    hemi : 'lh' | 'rh' | 'both' | 'split'
        'both': both hemispheres are shown in the same window;
        'split': hemispheres are displayed side-by-side in different viewing
        panes.
    surf : str
        Freesurfer surface mesh name (ie 'white', 'inflated', etc.).
    title : str
        Title for the window.
    cortex : str, tuple, dict, or None
        Specifies how the cortical surface is rendered. Options:

            1. The name of one of the preset cortex styles:
               ``'classic'`` (default), ``'high_contrast'``,
               ``'low_contrast'``, or ``'bone'``.
            2. A color-like argument to render the cortex as a single
               color, e.g. ``'red'`` or ``(0.1, 0.4, 1.)``. Setting
               this to ``None`` is equivalent to ``(0.5, 0.5, 0.5)``.
            3. The name of a colormap used to render binarized
               curvature values, e.g., ``Grays``.
            4. A list of colors used to render binarized curvature
               values. Only the first and last colors are used. E.g.,
               ['red', 'blue'] or [(1, 0, 0), (0, 0, 1)].
            5. A container with four entries for colormap (string
               specifiying the name of a colormap), vmin (float
               specifying the minimum value for the colormap), vmax
               (float specifying the maximum value for the colormap),
               and reverse (bool specifying whether the colormap
               should be reversed. E.g., ``('Greys', -1, 2, False)``.
            6. A dict of keyword arguments that is passed on to the
               call to surface.
    alpha : float in [0, 1]
        Alpha level to control opacity of the cortical surface.
    background, foreground : matplotlib colors
        color of the background and foreground of the display window
    subjects_dir : str | None
        If not None, this directory will be used as the subjects directory
        instead of the value set using the SUBJECTS_DIR environment
        variable.
    views : list | str
        views to use
    offset : bool
        If True, aligs origin with medial wall. Useful for viewing inflated
        surface where hemispheres typically overlap (Default: True)
    show_toolbar : bool
        If True, toolbars will be shown for each view.
    offscreen : bool
        If True, rendering will be done offscreen (not shown). Useful
        mostly for generating images or screenshots, but can be buggy.
        Use at your own risk.
    interaction : str
        Can be "trackball" (default) or "terrain", i.e. a turntable-style
        camera.
    w, h : int
        Figure width and height.
    axw, axh : int
        Width and height of the individual viewing panes.
    name : str
        Window title (alternative to ``title`` for consistency with other
        Eelbrain figures).
    pos : tuple of int
        Position of the new window on the screen.
    show : bool
        Currently meaningless due to limitation in VTK that does not allow
        hidden plots.
    run : bool
        Run the Eelbrain GUI app (default is True for interactive plotting and
        False in scripts).

    Notes
    -----
    The documentation lists only the methods that Eelbrain adds to or overrides
    from the PySurfer :class:`~surfer.Brain` super-class. For complete PySurfer
    functionality see te PySurfer documentation.
    """
    _display_time_in_frame_title = True

    def __init__(self, subject, hemi, surf='inflated', title=None,
                 cortex="classic", alpha=1.0, background="white",
                 foreground="black", subjects_dir=None, views='lat',
                 offset=True, show_toolbar=False, offscreen=False,
                 interaction='trackball', w=None, h=None, axw=None, axh=None,
                 name=None, pos=None, source_space=None, show=True, run=None):
        if not SURFER_IMPORTED:
            raise RuntimeError("PySurfer import failed. You should have seen a warning 'Error importing PySurfer' earlier.")

        from ._wx_brain import BrainFrame

        self.__data = []
        self.__annot = None
        self.__labels = {}  # {name: color}
        self.__time_index = 0
        self.__source_space = source_space
        self.hemi = hemi

        if isinstance(views, str):
            views = [views]
        elif not isinstance(views, list):
            views = list(views)
        n_rows = len(views)

        if hemi == 'split':
            n_columns = 2
        elif hemi in ('lh', 'rh', 'both'):
            n_columns = 1
        else:
            raise ValueError(f"{hemi=}")

        # layout
        if w is None and axw is None:
            if h is None and axh is None:
                axw = BRAIN_W
                axh = BRAIN_H
            else:
                if axh is None:
                    axh = int(round(h / n_rows))
                axw = int(round(axh * (BRAIN_W / BRAIN_H)))
        elif h is None and axh is None:
            if axw is None:
                axw = int(round(w / n_columns))
            axh = int(round(axw * (BRAIN_H / BRAIN_W)))

        if w is None:
            w = axw * n_columns

        if h is None:
            h = axh * n_rows

        if title is None:
            if name is None:
                title = subject
            elif isinstance(name, str):
                title = name
            else:
                raise TypeError(f"{name=} (str required)")
        elif not isinstance(title, str):
            raise TypeError(f"{title=} (str required)")

        if foreground is None:
            foreground = 'black'
        if background is None:
            background = 'white'

        if INLINE_DISPLAY:
            self._frame = figure = None
            self._png_repr_meta = {'height': h, 'width': w}
        else:
            self._frame = BrainFrame(None, self, title, w, h, n_rows, n_columns, surf, pos)
            figure = self._frame.figure
            self._png_repr_meta = None

        if subjects_dir is not None:
            subjects_dir = os.path.expanduser(subjects_dir)
        surfer.Brain.__init__(self, subject, hemi, surf, '', cortex, alpha, (w, h), background, foreground, figure, subjects_dir, views, offset, show_toolbar, offscreen, interaction)
        TimeSlicer.__init__(self)

        if self._frame and CONFIG['show'] and show:
            self._frame.Show()
            if CONFIG['eelbrain'] and do_autorun(run):
                from .._wxgui import run as run_gui  # lazy import for docs
                run_gui()

    def __repr__(self):
        args = [self.subject_id, self._hemi, self.surf]
        if self.n_times:
            args.append(f"{self.n_times} time points {ms(self._time_dim.tmin)}-{ms(self._time_dim.tstop)} ms")
        return f"<plot.brain.Brain: {', '.join(args)}>"

    def _asfmtext(self, **_):
        return self.image()

    # pysurfer 0.10
    def _ipython_display_(self):
        """Called by Jupyter notebook to display a brain."""
        if use_inline_backend():
            import IPython.display
            IPython.display.display(self.image())
        else:
            print(repr(self))

    def _check_source_space(self, source):
        "Make sure SourceSpace is compatible"
        source = get_source_dim(source)
        if source.subject != self.subject_id:
            raise ValueError(f"Trying to plot NDVar with {source.subject=} on Brain with subject={self.subject_id!r}")
        elif self._hemi == 'lh' and source.lh_n == 0:
            raise ValueError("Trying to add NDVar without lh data to plot of lh")
        elif self._hemi == 'rh' and source.rh_n == 0:
            raise ValueError("Trying to add NDVar without rh data to plot of rh")
        return source

    def add_label(self, label, color=None, alpha=1, scalar_thresh=None, borders=False, hemi=None, subdir=None, lighting=False, **kwargs):
        surfer.Brain.add_label(self, label, color, alpha, scalar_thresh,
                               borders, hemi, subdir)
        self.labels_dict[label.name][0].actor.property.lighting = lighting
        if color is None:
            color = getattr(label, 'color', None) or "crimson"
        name = label if isinstance(label, str) else label.name
        self.__labels[name] = color

    def add_mask(self, source, color=(0, 0, 0, 0.5), smoothing_steps=None,
                 alpha=None, subjects_dir=None):
        """Add a mask shading areas that are not included in an NDVar

        Parameters
        ----------
        source : SourceSpace
            SourceSpace.
        color : matplotlib color
            Mask color, can include alpha (defauls is black with alpha=0.5:
            ``(0, 0, 0, 0.5)``).
        smoothing_steps : scalar (optional)
            Smooth transition at the mask's border. If smoothing, the mask is
            added as data layer, otherwise it is added as label.
        alpha : scalar
            Alpha for the mask (supercedes alpha in ``color``).
        subjects_dir : str
            Use this directory as the subjects directory.
        """
        source = self._check_source_space(source)
        if color is True:
            color = (0, 0, 0, 0.5)
        color = to_rgba(color, alpha)
        if smoothing_steps is not None:
            # generate LUT
            lut = np.repeat(np.reshape(color, (1, 4)), 256, 0)
            lut[:, 3] = np.linspace(color[-1], 0, 256)
            np.clip(lut, 0, 1, lut)
            lut *= 255
            lut = np.round(lut).astype(np.uint8)
            # generate mask Label
            mask_ndvar = source._mask_ndvar(subjects_dir)
            self.add_ndvar(mask_ndvar, lut, 0., 1., smoothing_steps, False, None, False)
        else:
            lh, rh = source._mask_label(subjects_dir)
            if self._hemi == 'lh':
                rh = None
            elif self._hemi == 'rh':
                lh = None

            if source.lh_n and lh:
                self.add_label(lh, color[:3], color[3])
                self.labels_dict['mask-lh'][0].actor.property.lighting = False
            if source.rh_n and rh:
                self.add_label(rh, color[:3], color[3])
                self.labels_dict['mask-rh'][0].actor.property.lighting = False

    def add_ndvar(
            self,
            ndvar: NDVar,
            cmap: Any = None,
            vmin: float = None,
            vmax: float = None,
            smoothing_steps: int = None,
            colorbar: bool = False,
            time_label: Union[str, Callable] = 'ms',
            lighting: bool = False,
            contours: Union[bool, Sequence[float]] = None,
            alpha: float = 1,
            remove_existing: bool = False,
    ):
        """Add data layer form an NDVar

        Parameters
        ----------
        ndvar : NDVar  ([case,] source[, time])
            NDVar with SourceSpace dimension and optional time dimension. If it
            contains a :class:`Case` dimension, the average over cases is
            displayed. Values outside of the source-space, as well as masked
            values are set to 0, assuming a colormap in which 0 is transparent.
        cmap : str | list of matplotlib colors | array
            Colormap. Can be the name of a matplotlib colormap, a list of
            colors, or a custom lookup table (an n x 4 array with RBGA values
            between 0 and 255).
        vmin
            Lower endpoint for the colormap. Needs to be set explicitly if
            ``cmap`` is a LUT array.
        vmax
            Upper endpoint for the colormap. Needs to be set explicitly if
            ``cmap`` is a LUT array.
        smoothing_steps
            Number of smoothing steps if data is spatially undersampled
            (PySurfer ``Brain.add_data()`` argument).
        colorbar
            Add a colorbar to the figure (use ``.plot_colorbar()`` to plot a
            colorbar separately).
        time_label
            Label to show time point. Use ``'ms'`` or ``'s'`` to display time in
            milliseconds or in seconds, or supply a custom formatter for time
            values in seconds (default is ``'ms'``).
        lighting
            The data overlay is affected by light sources (default ``False``, 
            i.e. data overlays appear luminescent).
        contours
            Draw contour lines instead of a solid overlay. Set to a list of
            contour levels or ``True`` for automatic contours.
        alpha
            Alpha value for the data layer (0 = tranparent, 1 = opaque).
        remove_existing
            Remove data layers that have been added previously (default False).
        """
        ndvar = asndvar(ndvar)
        # check input data and dimensions
        source = self._check_source_space(ndvar)
        # find ndvar time axis
        if ndvar.ndim == 1 + ndvar.has_case:
            if ndvar.has_case:
                ndvar = ndvar.mean('case')
            time_dim = times = None
            data_dims = (source.name,)
        elif ndvar.ndim != 2:
            if ndvar.has_case:
                raise ValueError(f"{ndvar=}: must be one- or two dimensional. If you meant to plot the average of cases, use ndvar.mean('case')")
            raise ValueError(f"{ndvar=}: must be one- or two dimensional")
        elif ndvar.has_dim('time'):
            time_dim = ndvar.time
            times = ndvar.time.times
            data_dims = (source.name, 'time')
            if time_label == 'ms':
                time_label = lambda x: '%s ms' % int(round(x * 1000))
            elif time_label == 's':
                time_label = '%.3f s'
            elif time_label is False:
                time_label = None
        else:
            data_dims = ndvar.get_dimnames((source.name, None))
            time_dim = ndvar.get_dim(data_dims[1])
            times = np.arange(len(time_dim))
            time_label = None

        # remove existing data before modifying attributes
        if remove_existing:
            self.remove_data()

        # make sure time axis is compatible with existing data
        if isinstance(time_dim, UTS):
            self._init_time_dim(time_dim)

        # find colormap parameters
        meas = ndvar.info.get('meas')
        if cmap is None or isinstance(cmap, (str, Colormap)):
            epochs = ((ndvar,),)
            cmaps = find_fig_cmaps(epochs, cmap, alpha=True)
            vlims = find_fig_vlims(epochs, vmax, vmin, cmaps)
            vmin, vmax = vlims[meas]
        # colormap
        if contours is not None:
            if cmap is None:
                cmap = ('w', 'w')
            elif isinstance(cmap, str) and len(cmap) > 1:
                cmap = cmaps[meas]
            else:
                contour_color = to_rgb(cmap)
                cmap = (contour_color, contour_color)
        elif cmap is None or isinstance(cmap, str):
            cmap = cmaps[meas]

        # general PySurfer data args
        if smoothing_steps is None and source.kind == 'ico':
            smoothing_steps = source.grade + 1

        # determine which hemi we're adding data to
        if self._hemi in ('lh', 'rh'):
            data_hemi = self._hemi
        elif not source.lh_n:
            data_hemi = 'rh'
        elif not source.rh_n:
            data_hemi = 'lh'
        else:
            data_hemi = 'both'
        # remember where to find data_dict
        dict_hemi = 'rh' if data_hemi == 'rh' else 'lh'
        data_index = len(self._data_dicts[dict_hemi])

        # add data
        new_surfaces = []
        if data_hemi != 'rh':
            if self._hemi == 'lh':
                colorbar_ = colorbar
                colorbar = False
                time_label_ = time_label
                time_label = None
            else:
                colorbar_ = False
                time_label_ = None

            src_hemi = ndvar.sub(**{source.name: 'lh'})
            data = src_hemi.get_data(data_dims, 0)
            vertices = source.lh_vertices
            self.add_data(data, vmin, vmax, None, cmap, alpha, vertices, smoothing_steps, times, time_label_, colorbar_, 'lh')
            new_surfaces.extend(self.data_dict['lh']['surfaces'])

        if data_hemi != 'lh':
            src_hemi = ndvar.sub(**{source.name: 'rh'})
            data = src_hemi.get_data(data_dims, 0)
            vertices = source.rh_vertices
            self.add_data(data, vmin, vmax, None, cmap, alpha, vertices, smoothing_steps, times, time_label, colorbar, 'rh')
            new_surfaces.extend(self.data_dict['rh']['surfaces'])

        # update surfaces
        for surface in new_surfaces:
            if contours is not None:
                surface.enable_contours = True
                # http://code.enthought.com/projects/files/ets_api/enthought.mayavi.components.contour.Contour.html
                surface.contour.auto_update_range = False
                # surface.contour.maximum_contour = ndvar.max()
                # surface.contour.minimum_contour = ndvar.min()
                if contours is not True:
                    surface.contour.contours = contours
                    surface.contour.auto_contours = False

            if not lighting:
                surface.actor.property.lighting = False

        self.__data.append({
            'hemi': data_hemi,
            'data': ndvar,
            'dict_hemi': dict_hemi,
            'dict_index': data_index,
            'cmap': cmap,
            'vmin': vmin,
            'vmax': vmax,
        })

    def add_ndvar_annotation(self, ndvar, colors=None, borders=True, alpha=1, lighting=True):
        """Add annotation from labels in an NDVar
        
        Parameters
        ----------
        ndvar : NDVar of int
            NDVar in which each unique integer indicates a label. By default, 
            ``0`` is interpreted as unlabeled, but this can be overridden by 
            providing a ``colors`` dictionary that contains an entry for ``0``.
        colors : dict
            Dictionary mapping label IDs to colors.
        borders : bool | int
            Show label borders (instead of solid labels). If int, specify the 
            border width.
        alpha : scalar [0, 1]
            Opacity of the labels (default 1).
        lighting : bool
            Labels are affected by lights (default True).
        """
        source = self._check_source_space(ndvar)
        x = ndvar.get_data(source.name)
        if x.dtype.kind not in 'iu':
            raise TypeError("Need NDVar of integer type, not %r" % (x.dtype,))
        # determine colors
        label_values = np.unique(x)
        if colors is None or isinstance(colors, str):
            cells = np.setdiff1d(label_values, [0], assume_unique=True)
            colors = plot_colors = colors_for_oneway(cells, cmap=colors)
        elif isinstance(colors, dict):
            missing = np.setdiff1d(label_values, colors)
            # if 0 should be plotted, we need to shift values
            if 0 in colors and 0 in label_values:
                plot_colors = {k+1: v for k, v in colors.items()}
                x = x + 1
                label_values += 1
            else:
                plot_colors = colors
                missing = np.setdiff1d(missing, [0], assume_unique=True)
            if missing:
                raise KeysMissing(missing, 'colors', colors)
        else:
            raise TypeError(f"colors={colors}")
        # generate color table
        ctab = np.zeros((len(label_values), 5), int)
        ctab[:, 4] = label_values
        for i, v in enumerate(label_values):
            if v == 0:
                continue
            ctab[i, :4] = [int(round(c * 255.)) for c in to_rgba(plot_colors[v])]
        # generate annotation
        sss = source.get_source_space()
        indexes = (slice(None, source.lh_n), slice(source.lh_n, None))
        annot = []
        has_annot = []
        for ss, vertices, index in zip(sss, source.vertices, indexes):
            hemi = HEMI_ID_TO_STR[ss['id']]
            if self._hemi == OTHER_HEMI[hemi]:
                continue
            # expand to full source space
            ss_map = np.zeros(ss['nuse'], int)
            ss_map[np.in1d(ss['vertno'], vertices)] = x[index]
            # select only the used colors; Mayavi resets the range of the data-
            # to-LUT mapping to the extrema of the data at various points, so it
            # is safer to restrict the LUT to used colors
            ctab_index = np.in1d(ctab[:, 4], ss_map)
            hemi_ctab = ctab[ctab_index]
            if np.any(hemi_ctab):
                # map nearest from vertex to index
                nearest = np.searchsorted(ss['vertno'], ss['nearest'])
                # expand to full brain
                full_map = ss_map[nearest]
                annot.append((full_map, hemi_ctab))
                has_annot.append(hemi)

        if len(annot) == 0:
            return
        elif len(annot) == 1:
            annot = annot[0]
            hemi = has_annot[0]
        else:
            hemi = None

        self.add_annotation(annot, borders, alpha, hemi)
        self.__annot = colors
        if not lighting:
            for annot in self.annot_list:
                annot['surface'].actor.property.lighting = False

    def add_ndvar_label(self, ndvar, color=(1, 0, 0), borders=False, name=None,
                        alpha=None, lighting=False):
        """Draw a boolean NDVar as label.

        Parameters
        ----------
        ndvar : NDVar
            Boolean NDVar.
        color : matplotlib-style color | None
            anything matplotlib accepts: string, RGB, hex, etc. (default
            "crimson")
        borders : bool | int
            Show only label borders. If int, specify the number of steps
            (away from the true border) along the cortical mesh to include
            as part of the border definition.
        name : str
            Name for the label (for display in legend).
        alpha : float in [0, 1]
            alpha level to control opacity
        lighting : bool
            Whether label should be affected by lighting (default False).

        Notes
        -----
        To remove previously added labels, run Brain.remove_labels(). This
        method can only plot static labels; to plot a contour that varies over
        time, use :meth:`Brain.add_ndvar` with the ``contours`` parameter.
        """
        source = self._check_source_space(ndvar)
        x = ndvar.get_data(source.name)
        if x.dtype.kind != 'b':
            raise ValueError("Require NDVar of type bool, got %r" % (x.dtype,))
        if name is None:
            name = str(ndvar.name)
        color = to_rgba(color, alpha)
        lh_vertices = source.lh_vertices[x[:source.lh_n]]
        rh_vertices = source.rh_vertices[x[source.lh_n:]]
        lh, rh = source._label((lh_vertices, rh_vertices), name, color[:3])
        if lh and self._hemi != 'rh':
            while lh.name in self.labels_dict:
                lh.name += '_'
            self.add_label(lh, color[:3], color[3], borders=borders)
            self.labels_dict[lh.name][0].actor.property.lighting = lighting
        if rh and self._hemi != 'lh':
            while rh.name in self.labels_dict:
                rh.name += '_'
            self.add_label(rh, color[:3], color[3], borders=borders)
            self.labels_dict[rh.name][0].actor.property.lighting = lighting
        self.__labels[name] = color

    def add_ndvar_p_map(self, p_map, param_map=None, p0=0.05, p1=0.01, p0alpha=0.5, *args, **kwargs):
        """Add a map of p-values as data-layer

        Parameters
        ----------
        p_map : NDVar | NDTest
            Map of p values, or test result.
        param_map : NDVar
            Statistical parameter covering the same data points as p_map. Only the
            sign is used, for incorporating the directionality of the effect into
            the plot.
        p0 : scalar
            Highest p-value that is visible.
        p1 : scalar
            P-value where the colormap changes from ramping alpha to ramping color.
        p0alpha : 1 >= float >= 0
            Alpha at ``p0``. Set to 0 for a smooth transition, or a larger value to
            clearly delineate significant regions (default 0.5).
        ...
            Other parameters for :meth:`.add_ndvar`.
        """
        from .._stats.testnd import NDTest, MultiEffectNDTest

        if isinstance(p_map, NDTest):
            if isinstance(p_map, MultiEffectNDTest):
                raise NotImplementedError(f"plot.brain.p_map for {p_map.__class__.__name__}")
            elif param_map is not None:
                raise TypeError(f"param_map={param_map!r} when p_map is NDTest result")
            res = p_map
            p_map = res.p
            param_map = res._statistic_map
        p_map, cmap, vmax = p_lut(p_map, param_map, p0, p1, p0alpha)
        self.add_ndvar(p_map, cmap, -vmax, vmax, *args, **kwargs)

    def close(self):
        "Close the figure window"
        if self._frame:
            self._frame.Close()
        else:
            self._surfer_close()

    @property
    def closed(self):
        return self._figures[0][0] is None

    def _surfer_close(self):
        surfer.Brain.close(self)

    def copy_screenshot(self):
        "Copy the currently shown image to the clipboard"
        from .._wxgui import wx

        tempdir = mkdtemp()
        tempfile = os.path.join(tempdir, "brain.png")
        self.save_image(tempfile, 'rgba', True)

        bitmap = wx.Bitmap(tempfile, wx.BITMAP_TYPE_PNG)
        bitmap_obj = wx.BitmapDataObject(bitmap)

        if not wx.TheClipboard.IsOpened():
            open_success = wx.TheClipboard.Open()
            if open_success:
                wx.TheClipboard.SetData(bitmap_obj)
                wx.TheClipboard.Close()
                wx.TheClipboard.Flush()

    def _get_cmap_params(self, layer=0, label=True):
        """Return parameters required to plot a colorbar"""
        data = self.__data[layer]
        data_dict = self._data_dicts[data['dict_hemi']][data['dict_index']]
        colormap = ListedColormap(data_dict['orig_ctable'] / 255., label)
        return colormap, data_dict['fmin'], data_dict['fmax']

    def _has_annot(self):
        return bool(self.__annot)

    def _has_data(self):
        return bool(self.__data)

    def _has_labels(self):
        return bool(self.__labels)

    def enable_vertex_selection(self, color='red'):
        """Find source space vertice by right-clicking on the brain

        After enabling this functionality, each right-click on the brain will
        mark the closest vertex and the vertex number will be printed in the
        terminal.

        Parameters
        ----------
        color : mayavi color
            Color for the vertex marker.

        Examples
        --------
        Load a source space and plot it to be able to select vertices::

            ss = SourceSpace.from_file('directory/mri_subjects', 'fsaverage', 'ico-4')
            brain = plot.brain.brain(ss)
            brain.enable_vertex_selection()

        """
        if self.__source_space is None:
            raise RuntimeError("Can't enable vertex selection for brian without source space")
        for brain in self.brains:
            func = partial(self._select_nearest_source, hemi=brain.hemi, color=color)
            brain._f.on_mouse_pick(func, button="Right")
            
    @cached_property
    def _tris_lh(self):
        return self.__source_space._read_surf('lh')[1]

    @cached_property
    def _tris_rh(self):
        return self.__source_space._read_surf('rh')[1]

    def _select_nearest_source(self, vertex, hemi, color='red'):
        if not isinstance(vertex, int):
            vertex = vertex.point_id

        ss_vertices = self.__source_space.vertices[hemi == 'rh']
        if vertex not in ss_vertices:
            tris = self._tris_lh if hemi == 'lh' else self._tris_rh
            selected = np.any(tris == vertex, 1)
            for i in range(7):
                selected_v = np.unique(tris[selected])
                index = np.in1d(ss_vertices, selected_v)
                if index.any():
                    vertex = ss_vertices[index][0]
                    break
                for v in selected_v:
                    selected |= np.any(tris == v, 1)
            else:
                print("No vertex found in 7 iterations")

        if 'selection' in self.foci_dict:
            self.foci_dict.pop('selection')[0].remove()
        self.add_foci([vertex], True, hemi=hemi, scale_factor=0.5, name='selection', color=color)
        tag = 'L' if hemi == 'lh' else 'R'
        print(f'{tag}{vertex}')

    def image(self, name=None, format='png', alt=None, mode='rgb'):
        """Create an FMText Image from a screenshot

        Parameters
        ----------
        name : str
            Name for the file (without extension; default is ``data.name`` or
            'brain').
        format : str
            File format (default 'png').
        alt : None | str
            Alternate text, placeholder in case the image can not be found
            (HTML `alt` tag).
        mode : ``'rgb'`` | ``'rgba'``
            ``'rgb'`` to render solid background, or ``'rgba'`` to
            include alpha channel for a transparent background (default).
        """
        if name is None:
            for data in self.__data:
                name = data['data'].name
                if name:
                    break
            else:
                name = 'brain'
        im = self.screenshot(mode, True)
        if self._png_repr_meta is None:
            w, h = self._frame.GetClientSize()
            meta = {'height': h, 'width': w}
        else:
            meta = self._png_repr_meta
        return Image.from_array(im, name, format, alt, **meta)

    def _im_array(self):
        im = self.screenshot('rgba', True)
        im *= 255
        return im.astype(np.int8)

    def plot_colorbar(self, label=True, label_position=None, label_rotation=None,
                      clipmin=None, clipmax=None, orientation='horizontal',
                      width=None, ticks=None, layer=None, *args, **kwargs):
        """Plot a colorbar corresponding to the displayed data

        Parameters
        ----------
        label : str | bool
            Label for the x-axis (default is based on the data).
        label_position : 'left' | 'right' | 'top' | 'bottom'
            Position of the axis label. Valid values depend on orientation.
        label_rotation : scalar
            Angle of the label in degrees (For horizontal colorbars, the default is
            0; for vertical colorbars, the default is 0 for labels of 3 characters
            and shorter, and 90 for longer labels).
        clipmin : scalar
            Clip the color-bar below this value.
        clipmax : scalar
            Clip the color-bar above this value.
        orientation : 'horizontal' | 'vertical'
            Orientation of the bar (default is horizontal).
        width : scalar
            Width of the color-bar in inches.
        ticks : {float: str} dict | sequence of float
            Customize tick-labels on the colormap; either a dictionary with
            tick-locations and labels, or a sequence of tick locations.
        layer : int
            If the brain contains multiple data layers, plot a colorbar for
            only one (int in the order ndvars were added; default is to plot
            colorbars for all layers).

        Returns
        -------
        colorbar : :class:`~eelbrain.plot.ColorBar` | list
            ColorBar plot object (list of colorbars if more than one data layer
            are present).
        """
        if not self.__data:
            raise RuntimeError("Brain has no data to plot colorbar for")

        if layer is None:
            layers = range(len(self.__data))
        else:
            layers = (layer,)

        out = []
        for layer in layers:
            data = self.__data[layer]
            ndvar = data['data']
            if ticks is None:
                ticks = ndvar.info.get('cmap ticks')
            scale = AxisScale(ndvar, label)
            colormap, vmin, vmax = self._get_cmap_params(layer, label)
            out.append(ColorBar(colormap, vmin, vmax, label, label_position, label_rotation, clipmin, clipmax, orientation, scale, (), width, ticks, *args, **kwargs))
            # reset parames
            label = True

        if len(out) == 1:
            return out[0]
        else:
            return out

    def _set_annot(self, annot, borders, alpha):
        "Store annot name to enable plot_legend()"
        self.add_annotation(annot, borders, alpha)
        self.__annot = annot

    def play_movie(self, tstart=None, tstop=None, fps=10):
        """Play an animation by setting time"""
        max_wait = 1 / fps
        start = 0 if tstart is None else self.index_for_time(tstart)
        stop = self.n_times if tstop is None else self.index_for_time(tstop)
        for i in range(start, stop):
            t0 = time()
            self.set_data_time_index(i)
            dt = time() - t0
            if dt < max_wait:
                sleep(max_wait - dt)

    def plot_legend(self, *args, **kwargs):
        """Plot legend for parcellation

        Parameters
        ----------
        labels : dict (optional)
            Alternative (text) label for (brain) labels.
        h : 'auto' | scalar
            Height of the figure in inches. If 'auto' (default), the height is
            automatically increased to fit all labels.

        Returns
        -------
        legend : :class:`~eelbrain.plot.ColorList`
            Figure with legend for the parcellation.

        See Also
        --------
        plot.brain.annot_legend : plot a legend without plotting the brain
        """
        from ._brain import annot_legend
        if self.__labels:
            return ColorList(self.__labels, self.__labels.keys(), *args, **kwargs)
        elif self.__annot is None:
            raise RuntimeError("Can only plot legend for brain displaying "
                               "parcellation")
        elif isinstance(self.__annot, str):
            lh = os.path.join(self.subjects_dir, self.subject_id, 'label',
                              'lh.%s.annot' % self.__annot)
            rh = os.path.join(self.subjects_dir, self.subject_id, 'label',
                              'rh.%s.annot' % self.__annot)
            return annot_legend(lh, rh, *args, **kwargs)
        else:
            return ColorList(self.__annot, sorted(self.__annot), *args, **kwargs)

    def remove_data(self, hemi=None):
        """Remove data shown with ``Brain.add_ndvar``"""
        surfer.Brain.remove_data(self, None)
        del self.__data[:]
        self._time_dim = None

    def remove_labels(
            self,
            labels: Sequence[str] = None,
            mask: bool = False,
    ):
        """Remove labels shown with ``Brain.add_ndvar_label``

        Parameters
        ----------
        labels
            Labels to remove.
        mask
            Also remove the mask (labels with names starting with ``mask-``).
        """
        if labels is None and not mask:
            labels = [l for l in self.__labels if not l.startswith('mask-')]
        elif isinstance(labels, str):
            labels = [labels]
        else:
            labels = list(labels)
            if not all(isinstance(l, str) for l in labels):
                raise TypeError("labels=%r" % (labels,))
        surfer_labels = set(labels).intersection(self._label_dicts)
        surfer.Brain.remove_labels(self, surfer_labels)
        for label in labels:
            del self.__labels[label]

    def save_image(
            self,
            filename: PathArg,
            mode: Literal['rgb', 'rgba'] = 'rgb',
            antialiased: bool = False,
            fake_transparency: Any = None,
    ):
        """Save view from all panels to disk

        Parameters
        ----------
        filename: string
            Path to new image file.
        mode : ``'rgb'`` | ``'rgba'``
            ``'rgb'`` to render solid background (default), or ``'rgba'`` to
            include alpha channel for a transparent background.
        antialiased : bool
            Antialias the image (see :func:`mayavi.mlab.screenshot`
            for details; default False).

            .. warning::
               Antialiasing can interfere with ``rgba`` mode, leading to opaque
               background.

        fake_transparency
            Use this color as background color and make it transparent.
            Workaround if ``mode='rgba'`` is broken.

        See also
        --------
        .screenshot : grab current image as array

        Notes
        -----
        Due to limitations in TraitsUI, if multiple views or hemi='split'
        is used, there is no guarantee painting of the windows will
        complete before control is returned to the command line. Thus
        we strongly recommend using only one figure window (which uses
        a Mayavi figure to plot instead of TraitsUI) if you intend to
        script plotting commands.
        """
        if fake_transparency is not None:
            bg_color = to_rgb(fake_transparency)
            for figures in self._figures:
                for figure in figures:
                    figure.scene.background = bg_color
        else:
            bg_color = None
        im: np.ndarray = self.screenshot(mode, antialiased)
        if bg_color is not None:
            seed = np.zeros(im.shape[:2], bool)
            seed[0, 0] = True
            seed[0, -1] = True
            seed[-1, 0] = True
            seed[-1, -1] = True
            mask = np.all(im[:, :, :3] == [[bg_color]], 2)
            transparent = scipy.ndimage.binary_propagation(seed, mask=mask)
            im[:, :, 3] = ~transparent
        imsave(filename, im)

    def set_parallel_view(self, forward=None, up=None, scale=None):
        """Set view to parallel projection

        Parameters
        ----------
        forward : scalar
            Move the view forward (mm).
        up : scalar
            Move the view upward (mm).
        scale : scalar
            Mayavi parallel_scale parameter. Default is 95 for the inflated
            surface, 75 otherwise. Smaller numbers correspond to zooming in.
        """
        from mayavi import mlab
        from traits.trait_base import ETSConfig

        if scale is True:
            surf = self.geo['rh' if self._hemi == 'rh' else 'lh'].surf
            if surf == 'inflated':
                scale = 95 if ETSConfig.toolkit == 'wx' else 115
            else:
                scale = 75

        i = 0
        for figs in self._figures:
            for fig in figs:
                if forward is not None or up is not None:
                    mlab.view(focalpoint=(0, forward or 0, up or 0), figure=fig)
                if scale is not None:
                    if IS_OSX and self._frame and i == 0:
                        fig.scene.camera.parallel_scale = 2 * scale
                    else:
                        fig.scene.camera.parallel_scale = scale
                fig.scene.camera.parallel_projection = True
                fig.render()
                i += 1

    def set_size(self, width, height):
        """Set image size in pixels"""
        self._frame.SetImageSize(width, height)

    def set_surf(self, surf):
        from ._wx_brain import SURFACES

        self._set_surf(surf)
        if self._frame and surf in SURFACES:
            self._frame._surf_selector.SetSelection(SURFACES.index(surf))

    def _set_surf(self, surf):
        surfer.Brain.set_surf(self, surf)
        self.set_parallel_view(scale=True)

    def set_title(self, title):
        "Set the window title"
        self._frame.SetTitle(str(title))

    def set_vlim(self, v=None, vmax=None):
        """Change the colormap limits

        If the limit is symmetric, use ``set_vlim(vlim)``; if it is not, use
        ``set_vlim(vmin, vmax)``.

        Parameters
        ----------
        v : scalar
            If this is the only value specified it is interpreted as the upper
            end of the scale, and the lower end is determined based on
            the colormap to be ``-v`` or ``0``. If ``vmax`` is also specified,
            ``v`` specifies the lower end of the scale.
        vmax : scalar (optional)
            Upper end of the color scale.

        Notes
        -----
        Only affects the most recently added data layer.
        """
        if self.__data:
            data = self.__data[-1]
            cmap = data['cmap']
            if vmax is None:
                vmin, vmax = fix_vlim_for_cmap(None, abs(v), cmap)
            else:
                vmin = v
        else:
            data = None
            if vmax is None:
                surfer_data = self.data_dict.values()[0]
                vmax = abs(v)
                if surfer_data['fmin'] >= 0:
                    vmin = surfer_data['fmin']
                else:
                    vmin = -vmax
            else:
                vmin = v
        self.scale_data_colormap(vmin, (vmin + vmax) / 2, vmax, False)
        if data is not None:
            data['vmin'] = vmin
            data['vmax'] = vmax

    def get_vlim(self):
        "``(vmin, vmax)`` for the most recently added data layer"
        if not self.__data:
            raise RuntimeError('No data added with Brain.add_ndvar()')
        data = self.__data[-1]
        return data['vmin'], data['vmax']

    def _update_time(self, t, fixate):
        index = self._time_dim._array_index(t)
        if index == self.__time_index:
            return
        elif self.closed:
            return
        self.set_data_time_index(index)
        self.__time_index = index


if hasattr(surfer.Brain, 'add_label'):
    Brain.add_label.__doc__ = surfer.Brain.add_label.__doc__  # py3



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from collections import defaultdict

import numpy as np

from ._data_obj import Datalist, Dataset
from ._ndvar import neighbor_correlation
from ._info import BAD_CHANNELS


def _out(out, epochs):
    if out is None:
        return Datalist([[] for _ in range(len(epochs))])
    elif len(out) != len(epochs):
        raise ValueError("out needs same length as epochs, got %i/%i" %
                         (len(out), len(epochs)))
    return out


def new_rejection_ds(ds):
    """Create a rejection Dataset from a Dataset with epochs"""
    out = Dataset(info={BAD_CHANNELS: [], 'epochs.selection': ds.info.get('epochs.selection')})
    out['trigger'] = ds['trigger']
    out[:, 'accept'] = True
    out[:, 'rej_tag'] = ''
    return out


def find_flat_epochs(epochs, flat=1e-13, out=None):
    out = _out(out, epochs)
    d = epochs.max('time') - epochs.min('time')
    for i, chi in zip(*np.nonzero(d.get_data(('case', 'sensor')) < flat)):
        ch = epochs.sensor.names[chi]
        if ch not in out[i]:
            out[i].append(ch)

    return out


def find_flat_evoked(epochs, flat=1e-14):
    average = epochs.mean('case')
    d = average.max('time') - average.min('time')
    return epochs.sensor.names[d < flat]


def find_noisy_channels(epochs, mincorr=0.35):
    names = epochs.sensor.names
    out_e = Datalist([list(names[neighbor_correlation(ep) < mincorr]) for ep in epochs])
    return out_e


def channel_listlist_to_dict(listlist):
    out = defaultdict(list)
    for i, chs in enumerate(listlist):
        for ch in chs:
            out[ch].append(i)
    return dict(out)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from .basic import (
    tqdm,
    PickleableDataClass,
    as_list, as_sequence, ask, deprecated, deprecated_attribute, deprecate_ds_arg, intervals,
    keydefaultdict, n_decimals, natsorted,
    log_level, set_log_level, ScreenHandler,
)
from .system import IS_OSX, IS_WINDOWS, user_activity, restore_main_spec



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
import pytest

from eelbrain._experiment import test_def
from eelbrain._experiment.definitions import DefinitionError, find_dependent_epochs, find_epoch_vars, find_epochs_vars


def test_find_epoch_vars():
    assert find_epoch_vars({'sel': "myvar == 'x'"}) == {'myvar'}
    assert find_epoch_vars({'post_baseline_trigger_shift': "myvar"}) == {'myvar'}

    epochs = {'a': {'sel': "vara == 'a'"},
              'b': {'sel': "logical_and(varb == 'b', varc == 'c')"},
              'sec': {'sel_epoch': 'a', 'sel': "svar == 's'"},
              'super': {'sub_epochs': ('a', 'b')}}
    assert find_epochs_vars(epochs) == {'a': {'vara'},
                                        'b': {'logical_and', 'varb', 'varc'},
                                        'sec': {'vara', 'svar'},
                                        'super': {'vara', 'logical_and', 'varb', 'varc'}}
    assert set(find_dependent_epochs('a', epochs)) == {'sec', 'super'}
    assert find_dependent_epochs('b', epochs) == ['super']
    assert find_dependent_epochs('sec', epochs) == []
    assert find_dependent_epochs('super', epochs) == []


def test_find_test_vars():
    none = set()
    # t-test
    test = test_def.TTestRelated('A', 'a', 'b')
    assert test._find_test_vars() == ({'A'}, none)
    # groups
    test = test_def.TTestIndependent('group', 'a', 'b')
    assert test._find_test_vars() == (none, {'a', 'b'})
    # within-ANOVA
    test = test_def.ANOVA('a * b * subject')
    assert test.model == 'a%b'
    assert test._find_test_vars() == ({'a', 'b'}, none)
    # between ANOVA
    with pytest.raises(DefinitionError):
        test_def.ANOVA('a*b*c')
    test = test_def.ANOVA('a*b*c', model='')
    assert test.model == ''
    assert test._find_test_vars() == ({'a', 'b', 'c'}, none)
    # mixed ANOVA
    test = test_def.ANOVA('A * GR * subject(GR)')
    assert test.model == 'A'
    assert test._find_test_vars() == ({'A', 'GR'}, none)
    # two-stage
    test = test_def.TwoStageTest("a + b + a*b", vars={'a': 'c * d', 'b': 'c * e'})
    assert test._find_test_vars() == ({'c', 'd', 'e'}, none)
    test = test_def.TwoStageTest("a + b + a*b", vars={'a': 'c * d', 'b': 'c * e', 'x': 'something * nonexistent'})
    assert test._find_test_vars() == ({'c', 'd', 'e'}, none)
    test = test_def.TwoStageTest("a + b + a*b", vars={'a': ('c%d', {}), 'b': ('c%e', {})})
    assert test._find_test_vars() == ({'c', 'd', 'e'}, none)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
# cython: language_level=3, boundscheck=False, wraparound=False, cdivision=True

from libc.math cimport exp, sqrt, M_PI
import numpy as np

cimport numpy as np

ctypedef np.float64_t FLOAT64


def gaussian_smoother(FLOAT64[:,:] dist, double std):
    """Create a gaussian smoothing matrix

    Parameters
    ----------
    dist : array (float)
        Distances; dist[i, j] should provide the distance for any vertex pair i,
         j. Distances < 0 indicate absence of a connection.
    std : float
        The standard deviation of the kernel.

    Returns
    -------
    kernel : array (float64)
        Gaussian smoothing kernel, with same shape as dist.
    """
    cdef:
        Py_ssize_t source, target
        Py_ssize_t n_vertices = len(dist)
        double weight, target_sum
        double a = 1. / (std * sqrt(2 * M_PI))
        FLOAT64[:,:] out = np.empty((n_vertices, n_vertices))

    if dist.shape[1] != n_vertices:
        raise ValueError(f"dist needs to be rectangular, got shape {dist.shape}")

    for target in range(n_vertices):
        target_sum = 0
        for source in range(n_vertices):
            if dist[target, source] < 0:
                out[target, source] = 0
            else:
                weight = a * exp(- (dist[target, source] / std) ** 2 / 2)
                out[target, source] = weight
                target_sum += weight

        # normalize values for each target
        for source in range(n_vertices):
            out[target, source] /= target_sum

    return np.asarray(out)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from distutils.version import LooseVersion

import mne

MNE_EPOCHS = mne.BaseEpochs
MNE_RAW = mne.io.BaseRaw
MNE_EVOKED = mne.Evoked
MNE_LABEL = (mne.Label, mne.label.BiHemiLabel)

# VolVectorSourceEstimate added for 0.18
if LooseVersion(mne.__version__) >= LooseVersion('0.18'):
    MNE_VOLUME_STC = (mne.VolSourceEstimate, mne.VolVectorSourceEstimate)
else:
    MNE_VOLUME_STC = mne.VolSourceEstimate



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""Eelbrain-specific wx IDs"""
import wx

APYDOC_NEW = wx.NewIdRef()
ATTACH = wx.NewIdRef()

BASELINE_CUSTOM = wx.NewIdRef()
BASELINE_GLOABL_MEAN = wx.NewIdRef()
BASELINE_NONE = wx.NewIdRef()

CANVAS_PANEL_ZOOM = wx.NewIdRef()
CANVAS_PANEL_OVERVIEW = wx.NewIdRef()
CLEAR_CACHE = wx.NewIdRef()
CLEAR_TERMINAL = wx.NewIdRef()
COLOUR_CHOOSER = wx.NewIdRef()
COMMENT = wx.NewIdRef()
COPY_AS_PNG = wx.NewIdRef()

DATASET_ATTACH = wx.NewIdRef()
DATASET_IMPORT = wx.NewIdRef()
DRAW_CROSSHAIRS = wx.NewIdRef()
DUPLICATE = wx.NewIdRef()
DUPLICATE_WITH_OUTPUT = wx.NewIdRef()

EXPERIMENT_LOAD = wx.NewIdRef()
EXPERIMENT_NEW = wx.NewIdRef()
ENABLE_STARTUP_SCRIPT = wx.NewIdRef()
EXEC_DOCUMENT = wx.NewIdRef()
EXEC_DOCUMENT_FROM_DISK = wx.NewIdRef()
EXEC_FILE = wx.NewIdRef()
EXEC_SELECTION = wx.NewIdRef()

FIG_TITLE = wx.NewIdRef()
FOCUS_SHELL = wx.NewIdRef()

GRAND_AVERAGE = wx.NewIdRef()

HELP_EELBRAIN = wx.NewIdRef()
HELP_HIDE = wx.NewIdRef()
HELP_MDP = wx.NewIdRef()
HELP_MPL = wx.NewIdRef()
HELP_PDB = wx.NewIdRef()
HELP_PYTHON = wx.NewIdRef()

INSERT_Color = wx.NewIdRef()
INSERT_Path_file = wx.NewIdRef()
INSERT_Path_dir = wx.NewIdRef()
INSERT_Path_new = wx.NewIdRef()

LINK_TIME_AXES = wx.NewIdRef()

OPEN = wx.NewIdRef()
OPEN_HISTORY = wx.NewIdRef()
OPEN_HISTORY_CURRENT = wx.NewIdRef()
OPEN_HISTORY_PAST = wx.NewIdRef()

PLOT_COLORBAR = wx.NewIdRef()
PYDOC_NEW = wx.NewIdRef()
PYDOC_LOAD = wx.NewIdRef()
PYDOC_EXEC_ISOLATE = wx.NewIdRef()
PYDOC_HASH = wx.NewIdRef()
PYPLOT_CLOSEALL = wx.NewIdRef()
PYPLOT_DRAW = wx.NewIdRef()

P_MGR = wx.NewIdRef()
PCA_REMOVE = wx.NewIdRef()

RECENT_LOAD_E = wx.NewIdRef()
RECENT_LOAD_PYDOC = wx.NewIdRef()
REDO = wx.NewIdRef()
RESIZE = wx.NewIdRef()

SAVE_REJECTION = wx.NewIdRef()
SET_BAD_CHANNELS = wx.NewIdRef()
SET_FONT = wx.NewIdRef()
SET_LAYOUT = wx.NewIdRef()
SET_MARKED_CHANNELS = wx.NewIdRef()
SET_TEND = wx.NewIdRef()
SET_TITLE = wx.NewIdRef()
SET_TSTART = wx.NewIdRef()
SET_VLIM = wx.NewIdRef()
SHELL_Clear = wx.NewIdRef()
SHOW_EXAMPLES = wx.NewIdRef()
SHOW_SOURCES = wx.NewIdRef()
SIZE_MAX = wx.NewIdRef()
SIZE_MAX_NOTITLE = wx.NewIdRef()
SIZE_MIN = wx.NewIdRef()
SMOOTHING = wx.NewIdRef()
SPLITTER = wx.NewIdRef()
SPLITTER2 = wx.NewIdRef()

TABLE = wx.NewIdRef()
TABLE_COL = wx.NewIdRef()
TABLE_ROW = wx.NewIdRef()
TABLE_SAVE = wx.NewIdRef()
THRESHOLD = wx.NewIdRef()
TIME = wx.NewIdRef()
TOOLS = wx.NewIdRef()

UNDO = wx.NewIdRef()

VIEW_LATERAL = wx.NewIdRef()
VIEW_MEDIAL = wx.NewIdRef()

WINDOW_MINIMIZE = wx.NewIdRef()
WINDOW_TILE = wx.NewIdRef()
WINDOW_ZOOM = wx.NewIdRef()

YIELD_TO_TERMINAL = wx.NewIdRef()



#  Utilities working with (and needing to import) plots
from itertools import chain
import os
from typing import Dict, List, Tuple, Union

import numpy
from tqdm import tqdm

from .._types import PathArg
from .._utils import ui
from ._base import TimeSlicer


def save_movie(
        figures: Union[List[List[TimeSlicer]], Dict[Tuple[int, int], TimeSlicer]],
        filename: PathArg = None,
        time_dilation: float = 4,
        tstart: float = None,
        tstop: float = None,
        **kwargs,
):
    """Save a movie combining multiple figures with moving time axes

    Parameters
    ----------
    figures
        Either as a nested list of figures (each inner list is a row)
        or as a dictionary specifyig placement as ``{(xpos, ypos): figure}``.
    filename
        Filename for the movie (omit to use a GUI).
    time_dilation
        Factor by which to stretch time (default 4). Time dilation is
        controlled through the frame-rate; if the ``fps`` keyword argument
        is specified, ``time_dilation`` is ignored.
    tstart
        Time axis start (default is earliest time in ``figures``).
    tstop
        Time axis stop (default includes latest time in ``figures``).
    ...
        :func:`imageio.mimwrite` parmeters.

    Notes
    -----
    Might work best with::

        configure(prompt_toolkit=False)
    """
    import imageio
    from PIL import Image

    if filename is None:
        filename = ui.ask_saveas("Save movie...", "Save movie as...", [('Movie (*.mov)', '*.mov')])
        if not filename:
            return
    else:
        filename = os.path.expanduser(filename)

    # Interpret figures parameter
    if isinstance(figures, dict):
        figure_list = list(figures.values())
        coords = list(figures)
    elif isinstance(figures, (list, tuple)):
        if not isinstance(figures[0], (list, tuple)):
            figures = [figures]
        figure_list = list(chain.from_iterable(figures))
        coords = []
    else:
        raise TypeError(f"{figures=}")

    # Check time dimensions
    time_dims = list(filter(None, (getattr(f, '_time_dim', None) for f in figure_list)))
    if tstart is None:
        tstart = min(dim.tmin for dim in time_dims)
    if tstop is None:
        tstop = max(dim.tstop for dim in time_dims)
    if 'fps' in kwargs:
        tstep = time_dilation / kwargs['fps']
    else:
        tstep = min(dim.tstep for dim in time_dims)
        kwargs['fps'] = 1. / tstep / time_dilation

    # Generate movie
    times = numpy.arange(tstart, tstop, tstep)
    ims = []
    x_max = y_max = None
    for t in tqdm(times, "Frames", len(times)):
        t_ims = []
        for fig in figure_list:
            fig.set_time(t)
            f_im = Image.fromarray(fig._im_array(), 'RGBA')
            t_ims.append(f_im)

        # Infer positions
        if not coords:
            i = y = 0
            for row in figures:
                dy = 0
                x = 0
                for fig in row:
                    im = t_ims[i]
                    coords.append((x, y))
                    x += im.size[0]
                    dy = max(dy, im.size[1])
                    i += 1
                y += dy

        # Infer movie image dimensions
        if x_max is None:
            x_max = max(x + im.size[0] for (x, y), im in zip(coords, t_ims))
            y_max = max(y + im.size[1] for (x, y), im in zip(coords, t_ims))

        # Generate image
        im_buf = Image.new('RGBA', (x_max, y_max), (1, 1, 1, 0))
        for (x, y), im in zip(coords, t_ims):
            im_buf.paste(im, (x, y))

        # Store image
        ims.append(numpy.array(im_buf))

    # Save movie
    imageio.mimwrite(filename, ims, **kwargs)



# Author: Proloy Das <proloy@umd.edu>
"""2d projections of an ROI/mask image visualization via nilearn.plotting.glassbrain

Contains code from nilearn governed by the following license (3-Clause BSD):

Copyright (c) 2007 - 2015 The nilearn developers.
All rights reserved.


Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

  a. Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
  b. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  c. Neither the name of the nilearn developers nor the names of
     its contributors may be used to endorse or promote products
     derived from this software without specific prior written
     permission.


THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

"""
import os
from typing import Literal, Union
import warnings

import nibabel
import numpy as np

from .._data_obj import NDVarArg, Dataset, VolumeSourceSpace
from .._utils.numpy_utils import newaxis
from ._base import ColorBarMixin, TimeSlicerEF, Layout, EelFigure, brain_data, butterfly_data, use_inline_backend
from ._utsnd import Butterfly


# Copied from nilearn.image.resampling to avoid sklearn import with forced warnings
def get_bounds(shape, affine):
    adim, bdim, cdim = shape
    adim -= 1
    bdim -= 1
    cdim -= 1
    # form a collection of vectors for each 8 corners of the box
    box = np.array([[0.,   0,    0,    1],
                    [adim, 0,    0,    1],
                    [0,    bdim, 0,    1],
                    [0,    0,    cdim, 1],
                    [adim, bdim, 0,    1],
                    [adim, 0,    cdim, 1],
                    [0,    bdim, cdim, 1],
                    [adim, bdim, cdim, 1]]).T
    box = np.dot(affine, box)[:3]
    return list(zip(box.min(axis=-1), box.max(axis=-1)))


# Copied from nilearn.plotting.img_plotting
def _crop_colorbar( cbar, cbar_vmin, cbar_vmax ):
    """crop a colorbar to show from cbar_vmin to cbar_vmax.(symmetric_cbar=False)"""
    if (cbar_vmin is None) and (cbar_vmax is None):
        return
    cbar_tick_locs = cbar.locator.locs
    if cbar_vmax is None:
        cbar_vmax = cbar_tick_locs.max()
    if cbar_vmin is None:
        cbar_vmin = cbar_tick_locs.min()
    new_tick_locs = np.linspace(cbar_vmin, cbar_vmax,
                                len(cbar_tick_locs))
    cbar.ax.set_ylim(cbar.norm(cbar_vmin), cbar.norm(cbar_vmax))
    outline = cbar.outline.get_xy()
    outline[:2, 1] += cbar.norm(cbar_vmin)
    outline[2:6, 1] -= (1. - cbar.norm(cbar_vmax))
    outline[6:, 1] += cbar.norm(cbar_vmin)
    cbar.outline.set_xy(outline)
    cbar.set_ticks(new_tick_locs, update_ticks=True)


class GlassBrain(TimeSlicerEF, ColorBarMixin, EelFigure):
    """Plot 2d projections of a brain volume

    Based on :func:`nilearn.plotting.plot_glass_brain`.

    Parameters
    ----------
    ndvar
        Data to plot ([case,] time, source[, space]).
        If ``ndvar`` has a case dimension, the mean is plotted.
        If ``ndvar`` has a space dimension, the norm is plotted.
    cmap
        Colormap (name of a matplotlib colormap).
    vmin
        Plot data range minimum.
    vmax
        Plot data range maximum.
    dest
        If 'mri' the volume is defined in the coordinate system of
        the original T1 image. If 'surf' the coordinate system
        of the FreeSurfer surface is used (Surface RAS).
    mri_resolution
        If True the image is created in MRI resolution through upsampling.
        WARNING: it can result in significantly high memory usage.
    mni305
        Project data from MNI-305 space to MNI-152 space (by default this
        is enabled iff the source space subject is ``fsaverage``).
    black_bg
        If True, the background of the image is set to be black.
    display_mode
        Direction of the cuts:

        - ``'x'``: sagittal
        - ``'y'``: coronal
        - ``'z'``: axial
        - ``'l'``: sagittal, left hemisphere only
        - ``'r'``: sagittal, right hemisphere only
        - ``'ortho'``: three cuts in orthogonal directions, equivalent to
          ``'yxz'``

        Possible values are: 'ortho', 'x', 'y', 'z', 'xz', 'yx', 'yz',
        'l', 'r', 'lr', 'lzr', 'lyr', 'lzry', 'lyrz'. The default depends on
        the data: 'lyr' if both hemispheres are present, 'xz' if only one is.
    threshold
        If a number is given, values below the threshold (in absolute value) are
        plotted as transparent. If ``'auto'`` is given, the threshold is
        determined magically by analysis of the image.
    colorbar
        If True, display a colorbar on the right of the plots.
    draw_cross
        If draw_cross is True, a cross is drawn on the plot to
        indicate the cut plosition.
    annotate
        If annotate is True, positions and left/right annotation
        are added to the plot.
    alpha
        Alpha transparency for the brain schematics (between 0 and 1).
    plot_abs
        Plot the maximum intensity projection of the absolute value (rendering
        positive and negative values in the same manner). By default,
        (``False``), the sign of the maximum intensity will be represented with
        different colors. See `examples <http://nilearn.github.io/auto_examples/
        01_plotting/plot_demo_glass_brain_extensive.html>`_.
    draw_arrows
        Draw arrows in the direction of activation over the glassbrain plots.
        Naturally, for this to work ``ndvar`` needs to contain space dimension
        (i.e 3D vectors). By default it is set to ``True``.
    symmetric_cbar
        Specifies whether the colorbar should range from -vmax to vmax
        or from vmin to vmax. Setting to 'auto' will select the latter if
        the range of the whole image is either positive or negative.
        Note: The colormap will always be set to range from -vmax to vmax.
    interpolation
        Interpolation to use when resampling the image to the destination
        space. Can be "continuous" to use 3rd-order spline
        interpolation, or "nearest" to use nearest-neighbor mapping.
        "nearest" (default) is faster but can be noisier in some cases.
    show_time
        Add time label as text.
    title
        Figure title. Set to ``True`` to display current time point as figure
        title.
    ...
        Also accepts :ref:`general-layout-parameters`.

    Notes
    -----
    The brain overlay assumes coordinates in MNI152 space (see `The MNI brain
    and the Talairach atlas <http://imaging.mrc-cbu.cam.ac.uk/imaging/MniTalairach>`_).
    For data with different coordinates, the brain shape overlay will not be
    accurate.
    """
    _make_axes = False
    _display_time_in_frame_title = True

    def __init__(
            self,
            ndvar: NDVarArg,
            cmap: str = None,
            vmin: float = None,
            vmax: float = None,
            dest: Literal['mri', 'surf'] = 'mri',
            mri_resolution: bool = False,
            mni305: bool = None,
            black_bg: bool = False,
            display_mode: str = None,
            threshold: Union[float, Literal['auto']] = None,
            colorbar: bool = False,
            draw_cross: bool = True,
            annotate: bool = True,
            alpha: float = 0.7,
            plot_abs: bool = False,
            draw_arrows: bool = True,
            symmetric_cbar: Union[bool, Literal['auto']] = 'auto',
            interpolation: str = 'nearest',
            show_time: bool = False,
            data: Dataset = None,
            **kwargs):
        # Give wxPython a chance to initialize the menu before pyplot
        if not use_inline_backend():
            from .._wxgui import get_app
            get_app(jumpstart=True)

        # nilearn imports  matplotlib.pyplot and messes with the backend unless
        # DISPLAY is set
        old_display = os.environ.get('DISPLAY')
        if old_display is None:
            os.environ['DISPLAY'] = 'duh'
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', 'Trying to register the cmap', UserWarning)
            from nilearn.image import index_img
            from nilearn.plotting.displays import get_projector
            from nilearn.plotting.img_plotting import _get_colorbar_and_data_ranges
        if old_display is None:
            del os.environ['DISPLAY']

        # check parameters
        if vmax is not None and np.isnan(vmax):
            raise ValueError(f"{vmax=} (Tip: Use np.nanmax() instead of np.max())")
        elif vmin is not None and np.isnan(vmin):
            raise ValueError(f"{vmin=} (Tip: Use np.nanmin() instead of np.min())")

        if isinstance(ndvar, VolumeSourceSpace):
            source = ndvar
            ndvar = time = None
        else:
            ndvar = brain_data(ndvar, data)
            source = ndvar.get_dim('source')
            if not isinstance(source, VolumeSourceSpace):
                raise ValueError(f"{ndvar=}: need volume source space data")
            time = ndvar.get_dim('time') if ndvar.has_dim('time') else None
            if isinstance(ndvar.x, np.ma.MaskedArray) and np.all(ndvar.x.mask):
                ndvar = None

        if mni305 is None:
            mni305 = source.subject == 'fsaverage'

        if ndvar is None:
            cbar_vmin = cbar_vmax = imgs = img0 = dir_imgs = threshold = None
        else:
            if ndvar.has_case:
                ndvar = ndvar.mean('case')
            src = source.get_source_space()
            img = _stc_to_volume(ndvar, src, dest, mri_resolution, mni305)
            if time is not None:
                imgs = [index_img(img, i) for i in range(len(time))]
                img0 = imgs[0]
            else:
                img0 = img
                imgs = None
            if draw_arrows:
                if not ndvar.has_dim('space'):
                    draw_arrows = False
                    dir_imgs = None
                else:
                    dir_imgs = []
                    for direction in ndvar.space._directions:
                        dir_img = _stc_to_volume(ndvar.sub(space=direction), src, dest, mri_resolution, mni305)
                        if ndvar.has_dim('time'):
                            dir_imgs.append([index_img(dir_img, i) for i in range(len(ndvar.time))])
                        else:
                            dir_imgs.append([dir_img])
                    if plot_abs:
                        raise ValueError(f"Cannot use plot_abs={plot_abs} with draw_arrows={draw_arrows}")
            else:
                dir_imgs = None

            # determine parameters for colorbar
            if ndvar.has_dim('space'):
                data = ndvar.norm('space').x
            else:
                data = ndvar.x

            if cmap is None:
                if data.dtype.kind == 'b':
                    cmap = 'copper' if black_bg else 'Oranges'
                else:
                    cmap = 'cold_hot' if black_bg else 'cold_white_hot'

            if data.dtype.kind == 'b':
                if vmax is None:
                    vmax = 1
                if vmin is None:
                    vmin = 0
                cbar_vmin = cbar_vmax = None
            elif plot_abs:
                cbar_vmin, cbar_vmax, vmin, vmax = _get_colorbar_and_data_ranges(data, vmax, symmetric_cbar, (), 0)
            else:
                cbar_vmin, cbar_vmax, vmin, vmax = _get_colorbar_and_data_ranges(data, vmax, symmetric_cbar, ())

            # Deal with automatic settings of plot parameters
            if threshold == 'auto':
                # Threshold below a percentile value, to be sure that some
                # voxels pass the threshold
                threshold = _fast_abs_percentile(ndvar)
            if threshold is not None:
                threshold = float(threshold)
                if isinstance(ndvar.x, np.ma.MaskedArray):
                    raise ValueError(f"Cannot use threshold={threshold} with masked data")

        self.time = time
        self._ndvar = ndvar
        self._imgs = imgs
        self._dir_imgs = dir_imgs

        # layout
        if display_mode is None:
            has_lh = 'lh' in source.hemi
            has_rh = 'rh' in source.hemi
            if has_rh != has_lh:
                display_mode = 'xz'
            else:
                display_mode = 'lyr'
        n_plots = 3 if display_mode == 'ortho' else len(display_mode)
        layout = Layout(n_plots, 0.85, 2.6, tight=False, ncol=n_plots, **kwargs)
        # frame title
        if layout.name:
            frame_title = layout.name
        elif isinstance(layout.title, str):
            frame_title = layout.title
        elif ndvar is not None and ndvar.name:
            frame_title = ndvar.name
        else:
            frame_title = source.subject
        EelFigure.__init__(self, frame_title, layout)

        project = get_projector(display_mode)
        display = project(img0, alpha=alpha, plot_abs=plot_abs, threshold=threshold, figure=self.figure, axes=None, black_bg=black_bg, colorbar=colorbar)
        if img0:
            display.add_overlay(img0, threshold=threshold, interpolation=interpolation, colorbar=colorbar, vmin=vmin, vmax=vmax, cmap=cmap)

        ColorBarMixin.__init__(self, self._colorbar_params, ndvar)

        self.display = display
        self.threshold = threshold
        self.interpolation = interpolation
        self.cmap = cmap
        self.colorbar = colorbar
        self.vmin = vmin
        self.vmax = vmax
        self._arrows = draw_arrows

        if draw_arrows:
            if img0:
                self.arrow_scale = 7 * np.max(np.abs([vmax, vmin]))
                self._add_arrows(0)
        if annotate:
            display.annotate()
        if draw_cross:
            display.draw_cross()
        if hasattr(display, '_cbar'):
            cbar = display._cbar
            _crop_colorbar(cbar, cbar_vmin, cbar_vmax)

        if show_time:
            color = 'w' if black_bg else 'k'
            time_text = self.figure.text(0, 1, '', va='top', c=color, zorder=1000)
        else:
            time_text = None
        TimeSlicerEF.__init__(self, 'time', time, display_text=time_text)
        self._show()

    def _fill_toolbar(self, tb):
        ColorBarMixin._fill_toolbar(self, tb)

    # used by update_time
    def _add_arrows(self, t):
        """Adds arrows using matplotlib.quiver"""
        # Format 3D data
        data_list = []
        extent_list = []

        for display_ax in self.display.axes.values():
            data = []
            for k, direction in enumerate(self._ndvar.space._directions):
                vol = self._dir_imgs[k][t]
                try:
                    vol_data = np.squeeze(vol.get_data())
                    data_2d = display_ax.transform_to_2d(vol_data, vol.affine)
                    data_2d = np.squeeze(data_2d)
                except IndexError:
                    # We are cutting outside the indices of the data
                    data_2d = None
                data.append(data_2d)
            data_list.append(np.array(data))
            data_bounds = get_bounds(vol_data.shape, vol.affine)
            if display_ax.direction == 'y':
                (xmin, xmax), (_, _), (zmin, zmax) = data_bounds
            elif display_ax.direction in 'xlr':
                (_, _), (xmin, xmax), (zmin, zmax) = data_bounds
            elif display_ax.direction == 'z':
                (xmin, xmax), (zmin, zmax), (_, _) = data_bounds
            extent = (xmin, xmax, zmin, zmax)
            extent_list.append(extent)

        to_iterate_over = zip(self.display.axes.values(), data_list, extent_list)

        # Plotting using quiver
        if self.display._black_bg:
            color = 'w'
        else:
            color = 'k'
        ims = []
        for display_ax, data_2d, extent in to_iterate_over:
            if data_2d is not None:
                # get data mask
                if self.threshold is None:
                    thr = 0
                else:
                    thr = self.threshold ** 2
                data = (data_2d ** 2).sum(axis=0)
                not_mask = data > thr

                # If data_2d is completely masked, then there is nothing to
                # plot. Hence, continued to loop over. This problem came up
                # with matplotlib 2.1.0. See issue #9280 in matplotlib.
                if not_mask.any():
                    affine_2d = get_transform(extent, data.shape)
                    indices = np.where(not_mask)
                    x, y = coord_transform_2d(indices, affine_2d)
                    if display_ax.direction == 'y':
                        dir_data = (data_2d[0][indices], data_2d[2][indices])
                    elif display_ax.direction == 'l':
                        dir_data = (-data_2d[1][indices], data_2d[2][indices])
                    elif display_ax.direction in 'xr':
                        dir_data = (data_2d[1][indices], data_2d[2][indices])
                    elif display_ax.direction == 'z':
                        dir_data = (data_2d[0][indices], data_2d[1][indices])

                    im = display_ax.ax.quiver(x, y, dir_data[0], dir_data[1],
                                              color=color,
                                              scale=self.arrow_scale)
                else:
                    continue
            ims.append(im)

        self._quivers = ims

        return

    def _update_time(self, t, fixate):
        if self._ndvar is None:
            return
        index = self.time._array_index(t)
        # Redraw the overlay
        for axis in self.display._cut_displayed:
            for im in self.display.axes[axis].ax.images:
                im.remove()
        self.display.add_overlay(self._imgs[index], threshold=self.threshold, interpolation=self.interpolation, colorbar=False, vmin=self.vmin, vmax=self.vmax, cmap=self.cmap)
        # A little hack to make sure that the display still has correct colorbar flag.
        self.display._colorbar = self.colorbar

        # Redraw the arrows
        if self._arrows:
            for q in self._quivers:
                q.remove()
            self._add_arrows(index)

        self.canvas.redraw(self.figure.axes)

    def _colorbar_params(self):
        return self.cmap, self.vmin, self.vmax

    @classmethod
    def butterfly(cls, y, cmap=None, vmin=None, vmax=None, dest='mri',
                  mri_resolution=False, mni305=None, black_bg=False, display_mode=None,
                  threshold=None, colorbar=False, alpha=0.7, plot_abs=False,
                  draw_arrows=True, symmetric_cbar="auto", interpolation='nearest',
                  w=5, h=2.5, xlim=None, name=None, **kwargs):
        """Shortcut for a butterfly-plot with a time-linked glassbrain plot

        Parameters
        ----------
        y : NDVar  ([case,] time, source[, space])
            Data to plot; if ``ndvar`` has a case dimension, the mean is plotted.
            if ``ndvar`` has a space dimension, the norm is plotted.
        cmap : str
            Colormap (name of a matplotlib colormap).
        vmin : scalar
            Plot data range minimum.
        vmax : scalar
            Plot data range maximum.
        dest : 'mri' | 'surf'
            If 'mri' the volume is defined in the coordinate system of
            the original T1 image. If 'surf' the coordinate system
            of the FreeSurfer surface is used (Surface RAS).
        mri_resolution: bool, Default is False
            If True the image will be created in MRI resolution.
            WARNING: it can result in significantly high memory usage.
        mni305 : bool
            Project data from MNI-305 space to MNI-152 space (by default this
            is enabled iff the source space subject is ``fsaverage``).
        black_bg : boolean
            If True, the background of the image is set to be black.
        display_mode : str
            Direction of the cuts:

            - ``'x'``: sagittal
            - ``'y'``: coronal
            - ``'z'``: axial
            - ``'l'``: sagittal, left hemisphere only
            - ``'r'``: sagittal, right hemisphere only
            - ``'ortho'``: three cuts in orthogonal directions, equivalent to
              ``'yxz'``

            Possible values are: 'ortho', 'x', 'y', 'z', 'xz', 'yx', 'yz',
            'l', 'r', 'lr', 'lzr', 'lyr', 'lzry', 'lyrz'. Default depends on
            hemispheres in data.
        threshold : scalar | 'auto'
            If a number is given, values below the threshold (in absolute value) are
            plotted as transparent. If ``'auto'`` is given, the threshold is
            determined magically by analysis of the image.
        colorbar : boolean, Default is False
            If True, display a colorbar on the right of the plots.
        alpha : float between 0 and 1
            Alpha transparency for the brain schematics
        plot_abs : bool
            Plot the maximum intensity projection of the absolute value (rendering
            positive and negative values in the same manner). By default,
            (``False``), the sign of the maximum intensity will be represented with
            different colors. See `examples <http://nilearn.github.io/auto_examples/
            01_plotting/plot_demo_glass_brain_extensive.html>`_. Only affects
            GlassBrain plot.
        draw_arrows: boolean
            Draw arrows in the direction of activation over the glassbrain plots.
            Naturally, for this to work ``ndvar`` needs to contain space dimension
            (i.e 3D vectors). By default it is set to ``True``.
        symmetric_cbar : boolean or 'auto'
            Specifies whether the colorbar should range from -vmax to vmax
            or from vmin to vmax. Setting to 'auto' will select the latter if
            the range of the whole image is either positive or negative.
            Note: The colormap will always be set to range from -vmax to vmax.
        interpolation : str
            Interpolation to use when resampling the image to the destination
            space. Can be "continuous" (default) to use 3rd-order spline
            interpolation, or "nearest" to use nearest-neighbor mapping.
            "nearest" is faster but can be noisier in some cases.
        w : scalar
            Butterfly plot width (inches).
        h : scalar
            Plot height (inches; applies to butterfly and brain plot).
        xlim : scalar | (scalar, scalar)
            Initial x-axis view limits as ``(left, right)`` tuple or as ``length``
            scalar (default is the full x-axis in the data).
        name : str
            The window title (default is ndvar.name).

        Returns
        -------
        butterfly_plot : plot.Butterfly
            Butterfly plot.
        glassbrain : GlassBrain
            GlassBrain plot.
        """
        if use_inline_backend():
            jumpstart = False
        else:
            from .._wxgui import needs_jumpstart
            jumpstart = needs_jumpstart()

        hemis, bfly_data, brain_data = butterfly_data(y, None, return_vector_data=draw_arrows)

        if name is None:
            name = brain_data.name

        p = Butterfly(bfly_data, vmin=vmin, vmax=vmax, xlim=xlim, h=h, w=w, ncol=1, name=name, color='black', ylabel=hemis, axtitle=False)

        # Give wxPython a chance to initialize the menu before pyplot
        if jumpstart:
            from .._wxgui import get_app
            get_app().jumpstart()

        # GlassBrain plot
        p_glassbrain = GlassBrain(brain_data, cmap, vmin, vmax, dest, mri_resolution, mni305, black_bg, display_mode, threshold, colorbar, True, True, alpha, plot_abs, draw_arrows, symmetric_cbar, interpolation, h=h, name=name, **kwargs)

        p._auto_position(p_glassbrain)
        p.link_time_axis(p_glassbrain)

        return p, p_glassbrain


def _to_MNI152(trans):
    """Transfrom from MNI-305 space (fsaverage) to MNI-152

    parameters
    ----------
    trans: ndarray
        The affine transform.

    Notes
    -----
    uses approximate transformation mentioned `Link here <https://surfer.nmr.mgh.harvard.edu/fswiki/CoordinateSystems>`_
    """
    t = np.array([[ 0.9975, -0.0073,  0.0176, -0.0429],
                  [ 0.0146,  1.0009, -0.0024,  1.5496],
                  [-0.0130, -0.0093,  0.9971,  1.1840],
                  [ 0,       0,       0,       1     ]])
    return np.dot(t, trans)


def _stc_to_volume(ndvar, src, dest='mri', mri_resolution=False, mni305=False):
    """Save a volume source estimate in a NIfTI file.

    Parameters
    ----------
    ndvar : NDVar
        The source estimate
    src : list
        The list of source spaces (should actually be of length 1).
    dest : 'mri' | 'surf'
        If 'mri' the volume is defined in the coordinate system of
        the original T1 image. If 'surf' the coordinate system
        of the FreeSurfer surface is used (Surface RAS).
    mri_resolution : bool
        It True the image is saved in MRI resolution.
        WARNING: if you have many time points the file produced can be
        huge.
    mni305 : bool
        Set to True to convert RAS coordinates of a voxel in MNI305 space (fsaverage space)
        to MNI152 space via updating the affine transformation matrix.

    Returns
    -------
    img : instance Nifti1Image
        The image object.
    """
    src_type = src[0]['type']
    if src_type != 'vol':
        raise ValueError(f"You need a volume source space. Got type: {src_type}")
    if len(src) != 1:
        raise NotImplementedError("Multiple source spaces")

    if ndvar.has_dim('space'):
        ndvar = ndvar.norm('space')
    source = ndvar.get_dim('source')
    if ndvar.has_dim('time'):
        data = ndvar.get_data(('source', 'time'), mask=0)
    else:
        data = ndvar.get_data(('source', newaxis), mask=0)

    # check for infinite values and make them zero
    non_finite_mask = np.logical_not(np.isfinite(data))
    if non_finite_mask.sum() > 0:  # any non_finite_mask values?
        data[non_finite_mask] = 0

    # project data to 4d array
    n_times = data.shape[1]
    shape3d = src[0]['shape'][::-1]
    shape4d = (*shape3d, n_times)
    vol = np.zeros(shape4d)
    mask = src[0]['inuse'].astype(np.bool)
    if data.shape[0] < mask.sum():
        mask[mask] = np.in1d(src[0]['vertno'], source.vertices[0], assume_unique=True)
    vol[mask.reshape(shape3d)] = data
    vol = np.moveaxis(vol, 3, 0)  # time on first axis

    if mri_resolution:
        mri_shape3d = (src[0]['mri_height'], src[0]['mri_depth'],
                       src[0]['mri_width'])
        mri_shape = (n_times, src[0]['mri_height'], src[0]['mri_depth'],
                     src[0]['mri_width'])
        mri_vol = np.zeros(mri_shape)
        interpolator = src[0]['interpolator']
        for k, v in enumerate(vol):
            mri_vol[k] = (interpolator * v.ravel()).reshape(mri_shape3d)
        vol = mri_vol
        affine = src[0]['vox_mri_t']['trans']
    else:
        affine = src[0]['src_mri_t']['trans']

    if dest == 'mri':
        affine = np.dot(src[0]['mri_ras_t']['trans'], affine)
    elif dest == 'surf':
        affine = affine.astype(float)
    else:
        raise ValueError(f"dest={dest!r}")

    affine[:3] *= 1e3
    if mni305:
        affine = _to_MNI152(affine)

    # write the image in nifty format
    header = nibabel.nifti1.Nifti1Header()
    header.set_xyzt_units('mm', 'msec')
    if ndvar.has_dim('time'):
        header['pixdim'][4] = 1e3 * ndvar.time.tstep
    else:
        header['pixdim'][4] = None
    with warnings.catch_warnings(record=True):  # nibabel<->numpy warning
        img = nibabel.Nifti1Image(vol.T, affine, header=header)
    return img


def _fast_abs_percentile(ndvar, percentile=80):
    """A fast version of the percentile of the absolute value.

    Parameters
    ----------
    data: ndvar
        The input data
    percentile: number between 0 and 100
        The percentile that we are asking for

    Returns
    -------
    value: number
        The score at percentile

    Notes
    -----
    This is a faster, and less accurate version of
    scipy.stats.scoreatpercentile(np.abs(data), percentile)
    # inspired from nilearn._utils.extmath.fast_abs_percentile
    """
    if isinstance(ndvar.x, np.ma.masked_array):
        return
    if ndvar.has_dim('space'):
        data = ndvar.norm('space').x
    else:
        data = abs(ndvar.x)

    data = data.ravel()
    index = int(data.size * .01 * percentile)
    try:
        # Partial sort: faster than sort
        data = np.partition(data, index)
    except ImportError:
        data.sort()
    return data[index]


def get_transform(extent, shape):
    xmin, xmax, zmin, zmax = extent
    T = np.eye(3)
    T[0, 0] = (zmin - zmax) / shape[0]
    T[0, 2] = zmax
    T[1, 1] = (xmax - xmin) / shape[1]
    T[1, 2] = xmin

    return T


def coord_transform_2d(indices, affine):
    rows, cols = indices
    old_coords = np.array([np.array(rows), np.array(cols), np.ones(len(rows))])
    y, x, _ = np.dot(affine, old_coords)
    return x, y



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""Plot data split for cross-validation"""
from typing import Union

from matplotlib.patches import Rectangle
from matplotlib.ticker import MaxNLocator
import numpy

from .._trf.shared import Splits, split_data
from .._data_obj import Dataset, CategorialArg, NDVarArg, asndvar
from .._utils import deprecate_ds_arg
from ._base import EelFigure, Layout, LegendMixin, LegendArg
from ._styles import colors_for_oneway


class DataSplit(EelFigure, LegendMixin):

    def __init__(
            self,
            splits: Splits,
            legend: LegendArg = 'upper right',
            colors: dict = None,
            xlabel: str = 'Segment',
            **kwargs,
    ):
        """Plot data splits

        Parameters
        ----------
        splits
            The data splits.
        legend
            Matplotlib figure legend location argument or 'fig' to plot the
            legend in a separate figure.
        colors
            Customize colors (uses the following keys: ``'train', 'validate',
            'test'``).
        ...
            Also accepts :ref:`general-layout-parameters`.

        Notes
        -----
        This plot can also be generated from a :class:`BoostingResult` object
        through :meth:`BoostingResult.splits.plot` method.
        """
        attrs = ['train', 'validate', 'test']
        labels = {'train': 'Training', 'validate': 'Validation', 'test': 'Testing'}
        if colors is None:
            colors = colors_for_oneway(attrs, unambiguous=[6, 3, 5])

        h_default = max(2, 0.5 + 0.15 * len(splits.splits))
        layout = Layout(1, 16/9, h_default, **kwargs)
        EelFigure.__init__(self, None, layout)
        ax = self.figure.axes[0]

        for x in splits.segments[:-1, 1]:
            ax.axvline(x, color='k')
        handles = {}
        for y, split in enumerate(splits.splits):
            for attr in attrs:
                segments = getattr(split, attr)
                if segments is None:
                    continue
                rect = None
                for x0, x1 in segments:
                    rect = Rectangle((x0, y-0.5), x1-x0, 1, color=colors[attr])
                    ax.add_artist(rect)
                handles[attr] = rect
            # separate splits
            if y:
                ax.axhline(y-0.5, color='white')
        labels = {key: labels[key] for key in handles}
        ax.set_ylabel('Split')
        ax.set_ylim(-0.5, len(splits.splits)-0.5)
        ax.yaxis.set_major_locator(MaxNLocator(integer=True))
        if xlabel:
            ax.set_xlabel(xlabel)
        ax.set_xlim(splits.segments[0, 0], splits.segments[-1, 1])
        LegendMixin.__init__(self, legend, handles, labels)
        self._show()

    def _fill_toolbar(self, tb):
        LegendMixin._fill_toolbar(self, tb)


@deprecate_ds_arg
def preview_partitions(
    cases: Union[int, NDVarArg] = 0,
    partitions: int = None,
    model: CategorialArg = None,
    validate: int = 1,
    test: int = 0,
    data: Dataset = None,
    **kwargs,
) -> DataSplit:
    """Preview how data will be partitioned for the boosting function

    Parameters
    ----------
    cases
        Description of the data that will be used. Can be specified as
        :class:`NDVar` data, or as integer describing the number of trials
        (the default assumes continuous data).
    ...
        For a description of the splitting parameters see :func:`~eelbrain.boosting`.
        For plotting parameters see :class:`DataSplit`.

    See Also
    --------
    :ref:`exa-data_split` example
    """
    if isinstance(cases, int):
        if cases == 0:
            has_case = False
            if partitions is None:
                partitions = 2 + test + validate if test else 10
            ns = [partitions]
        else:
            has_case = True
            if partitions is None:
                raise NotImplementedError('Automatic partitions with trials')
            ns = [1] * cases
    else:
        y = asndvar(cases, data=data)
        has_case = y.has_case
        if y.has_case:
            n_cases = len(y)
        else:
            n_cases = 1
        ns = [len(y.get_dim('time'))] * n_cases
    index = numpy.cumsum([0] + ns)
    segments = numpy.hstack([index[:-1, None], index[1:, None]])
    splits = split_data(segments, partitions, model, data, validate, test)
    kwargs.setdefault('xlabel', 'Case' if has_case else 'Segment')
    return DataSplit(splits, **kwargs)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from ..fmtxt import Section
from ._sensor import sensor_results, sensor_time_results
from ._source import source_time_results
from ._uts import time_results


def result_report(res, ds, title=None, colors=None):
    """Automatically generate section from an NDTest result object

    Parameters
    ----------
    res : NDTest
        Test-result.
    ds : Dataset
        Dataset containing the data on which the test was performed.
    """
    sec = Section(title or res._name())

    dims = {dim.name for dim in res._dims}
    sec.append(res.info_list())

    if dims == {'time'}:
        sec.append(time_results(res, ds, colors))
    elif dims == {'sensor'}:
        sec.append(sensor_results(res, ds, colors))
    elif dims == {'time', 'sensor'}:
        sec.append(sensor_time_results(res, ds, colors))
    elif dims == {'time', 'source'}:
        sec.append(source_time_results(res, ds, colors))
    else:
        raise NotImplementedError("dims=%r" % dims)
    return sec
